from yo_fluq import Query, FileIO
from pathlib import Path
from foundation_kaia.releasing import MDHandler

def create_documentation():
    docu_root = Path(__file__).parent
    files = Query.folder(docu_root).where(lambda z: z.name.startswith('p')).order_by(lambda z: z.name).to_list()
    sections = []
    for file in files:
        text = FileIO.read_text(file)
        sections.extend(MDHandler.parse_code(text, True))


    template = FileIO.read_text(Path(__file__).parent/'readme_template.md')
    docu = MDHandler.write_sections(sections, 4)
    docu = template.replace('<<<FROM_TESTS>>>', docu)
    docu = "# Table of contents\n\n"+MDHandler.create_toc(docu) +'\n\n'+docu

    return docu





from ....framework import DockerWebServiceApi
import requests
from .settings import EspeakPhonemizerSettings
from .controller import EspeakPhonemizerController



class EspeakPhonemizer(DockerWebServiceApi[EspeakPhonemizerSettings, EspeakPhonemizerController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def phonemize(self, text: str|list[str], language: str = 'en-us', stress = False):
        if isinstance(text, str):
            text = [text]

        result = requests.post(
            self.endpoint('/echo'),
            json = dict(
                text = text,
                language = language,
                stress = stress,
            )
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint /phonemize returned unexpected status code {result.status_code}\n{result.text}")
        return result.json()

    Settings = EspeakPhonemizerSettings
    Controller = EspeakPhonemizerController



from abc import ABC, abstractmethod
from pathlib import Path

class IFileSystem(ABC):
    @abstractmethod
    def upload_file(self, local_path: Path, remote_path: str):
        pass

    @abstractmethod
    def download_file(self, remote_path: str, local_path: Path):
        pass

    @abstractmethod
    def create_empty_folder(self, path: str|Path):
        pass

    @abstractmethod
    def is_file(self, remote_path: str):
        pass

    @abstractmethod
    def is_dir(self, remote_path: str):
        pass

    @abstractmethod
    def delete_file_or_folder(self, remote_path: str):
        pass



from ...controllers import ControllerRegistry
from .operator_state import OperatorState
from .operator_log import OperatorLog
from sqlalchemy import Engine, select
from sqlalchemy.orm import Session
from .job import Job
import traceback
from abc import ABC, abstractmethod
from .trackable_session_factory import TrackableSessionFactory
from .job_for_planner import JobForPlanner
from ...common import Locator, Loc

class SessionFactory:
    def __init__(self, engine):
        self._engine = engine

    def __call__(self):
        return Session(self._engine)


class Core:
    def __init__(self,
                 engine: Engine,
                 registry: ControllerRegistry,
                 locator: Locator = Loc,
                 debug_output: bool = False,
                 ):
        self._engine = engine
        self.new_session = SessionFactory(self._engine)# TrackableSessionFactory(self._engine, Core.job_to_id)
        self.registry = registry
        self.locator = locator
        self.operator_states: dict[str, OperatorState] = dict()
        self.operator_log: OperatorLog = OperatorLog(debug_output)
        self.jobs_for_planner: tuple[JobForPlanner,...]|None = None

    @staticmethod
    def job_to_id(job: Job):
        return job.id

    def get_job_by_id(self, session: Session, id: str):
        return session.scalar(select(Job).where(Job.id == id))

    def close_job(self, session: Session, job: Job, custom_message: str):
        self.operator_log.task(job.id).event(custom_message)
        if custom_message is None:
            custom_message = traceback.format_exc()
        job.finished = True
        job.success = False
        job.error = custom_message


class ICoreAction(ABC):
    @abstractmethod
    def apply(self, core: Core):
        pass


from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase

def pure_http(test_case: TestCase, api: BrainBox.Api):
    """
    # Use it without Python

    All the BrainBox functionality is accessible without Python API, with direct HTTP requests.
    Since Python API actually works over the HTTP requests to BrainBox server,
    and the API calls are translated into HTTP requests automatically with `brainbox.framework.common.marshalling` module,
    I can guarantee that it works.
    However, it is not yet nicely documented.

    ## Manage the controllers

    Installation, running, stopping and self-testing of deciders is available on
    `http://127.0.0.1:8090/html/controllers/status`.
    The web-page is hopefully self-explanatory.

    The web-page also allows you to start the container.
    Since most of the containers have web-server inside, you can call the endpoints of these web-servers
    instead of BrainBox.
    These servers, however, are not nicely written, don't have the documentation
    and there is no plan to write this documentation in the future,
    because it's a massive work that will have little sense,
    since BrainBox implements the calls itself.

    ## Run the tasks

    To run the deciders via BrainBox, we have two endpoints, `/jobs/add` and `/jobs/join`.

    `jobs/add` accepts a list of dictionaries, each describing one job.
    What we did above with `BrainBox.Task.call(HelloBrainBox)` was actually a definition of the job with API.
    Job has the following fields:

    * `id`: a unique string id of the job
    * `decider`: the name of the decider, like `"Boilerlate"` string
    * `method`: the method of decider class we're running. Can be None if the class defines `__call__` method and is therefore callable itself.
    * `decider_parameter`: decider's parameter, usually None
    * `arguments`: arguments of the method as dictionary.
    * `info`: arbitrary data that is associated with the job, but doesn't have any effect on it.
    * `batch`: jobs in e.g. `Collector`'s packs are organized in batches, so in the web-interface it may be seen what percentage
       of this collective job is finished.
    * `dependencies`: a dictionary with keys set to the names of the arguments of the method, and values as id of the dependent tasks.
      If the key starts with `*`, the dependent task will be waited for, but it's value won't be used.
    * `ordering_token`: None or an arbitrary string that helps BrainBox arrange the jobs to minimize the model switching.
       Something like `<larger_model_name>/<smaller_model_name>/<even_smaller_model_name>` is usually sufficient.

    Simply the list of such dictionaries to the BrainBox server and it will add the tasks to the queue,
    returning the control immediately.
    To block your process until the tasks are finished, use `/jobs/join` endpoint,
    and provide it with the list of `id` of the jobs you want to wait for.
    As a result, `/jobs/join` will return a list of arbitrary objects that correspond to the list of `id`.

    """
    import requests
    import uuid

    address = api.address
    id = str(uuid.uuid4())

    reply = requests.post(f'http://{address}/jobs/add', json=
    {
        "arguments": {
            "jobs": [
                {
                    "id": id,
                    "decider": "HelloBrainBox",
                    "method": "json",
                    "decider_parameter": None,
                    "arguments": {
                        "argument": "Hello, HTTP"
                    }
                }
            ]
        }
    })
    if reply.status_code != 200:
        raise ValueError(f"Endpoint returned {reply.status_code}\n{reply.json()['error']}")

    reply = requests.get(f"http://{address}/jobs/join", json=
    {
        "arguments": {
            "ids": [
                id
            ]
        }
    })

    if reply.status_code != 200:
        raise ValueError(f"Endpoint returned {reply.status_code}\n{reply.json()['error']}")

    result = reply.json()['result'][0]
    test_case.assertDictEqual(
        {
            'argument': 'Hello, HTTP',
            'model': 'no_parameter',
            'setting': 'default_setting'
        },
        result
    )


from typing import *
from ...common import IDecider
from .service import BrainBoxService, BrainBoxServiceSettings
from ...controllers import ControllerRegistry, IController
from ...common import Locator, Loc


class ServerlessTest:
    def __init__(self,
                 services: Iterable[Union[IController, IDecider]],
                 time_limit_in_seconds: int = 10,
                 allow_failures: bool = False,
                 ):
        self.services = services
        self.time_limit_in_seconds = time_limit_in_seconds
        self.allow_failures = allow_failures
        self.folder = Loc.create_test_folder('brainbox_serverless_test_runs')


    def __enter__(self) -> BrainBoxService:
        self.folder.__enter__()
        locator = Locator(self.folder.path)
        self._service = BrainBoxService(BrainBoxServiceSettings(
            ControllerRegistry(self.services),
            locator=locator
        ))
        self._service.run()
        return self._service


    def __exit__(self, exc_type, exc_val, exc_tb):
        self._service.shutdown()
        self.folder.__exit__(exc_type, exc_val, exc_tb)







import os
from typing import Iterable
from unittest import TestCase

import requests

from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, INotebookableController, File
)
from ...common import check_if_its_sound, download_file
from .settings import OpenVoiceSettings
from pathlib import Path
import zipfile


class OpenVoiceController(
    DockerWebServiceController[OpenVoiceSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders={
                'pretrained/model/checkpoints' : '/home/app/checkpoints',
                'models' : '/models',
                'voices': '/voices',
                'temp': '/temp'
            },
        )

    def post_install(self):
        path = self.resource_folder('pretrained')/'checkpoints_1226.zip'
        download_file(
            'https://myshell-public-repo-host.s3.amazonaws.com/openvoice/checkpoints_1226.zip',
            path
        )
        if not zipfile.is_zipfile(path):
            raise zipfile.BadZipFile(f"The file {path} is not a valid zip file.")

        folder = self.resource_folder('pretrained','model')
        os.makedirs(folder, exist_ok=True)
        with zipfile.ZipFile(path, 'r') as zip_ref:
            zip_ref.extractall(folder)



    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return OpenVoiceSettings()

    def create_api(self):
        from .api import OpenVoice
        return OpenVoice()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import OpenVoice

        reference_speaker = File.read(Path(__file__).parent / "lina.wav")
        api.execute(BrainBoxTask.call(OpenVoice).train('test_voice', [reference_speaker]))
        yield TestReport.last_call(api).href('training').with_comment("Training a model for a reference voice")

        source_speaker = File.read(Path(__file__).parent/"nikita.wav")
        task = BrainBoxTask.call(OpenVoice).generate(source_speaker, 'test_voice')
        result_file = api.execute(task)

        check_if_its_sound(api.open_file(result_file).content, tc)
        yield TestReport.last_call(api).href('href').result_is_file(File.Kind.Audio)



DOCKERFILE = f'''

FROM python:3.9.21-slim

{SmallImageBuilder.APT_INSTALL('git unzip build-essential ffmpeg wget')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

WORKDIR /home/app

{SmallImageBuilder.GIT_CLONE_AND_RESET(
    'https://github.com/myshell-ai/OpenVoice.git',
    'OpenVoice',
    'bb79fa78a5a7a7a3d7602b7f6d48705213a039c7',
)}

COPY . /home/app/OpenVoice/

ENTRYPOINT ["python3", "/home/app/OpenVoice/main.py"]
'''

DEPENDENCIES = '''
aiofiles==23.2.1
altair==5.5.0
annotated-types==0.7.0
anyio==4.8.0
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==3.0.0
async-lru==2.0.4
attrs==25.1.0
audioread==3.0.1
av==10.0.0
babel==2.17.0
beautifulsoup4==4.13.1
bleach==6.2.0
blinker==1.9.0
certifi==2025.1.31
cffi==1.17.1
charset-normalizer==3.4.1
click==8.1.8
cn2an==0.5.22
coloredlogs==15.0.1
comm==0.2.2
contourpy==1.2.1
ctranslate2==3.24.0
cycler==0.12.1
Cython==3.0.11
debugpy==1.8.12
decorator==5.1.1
defusedxml==0.7.1
distro==1.9.0
dtw-python==1.4.4
eng-to-ipa==0.0.2
exceptiongroup==1.2.2
executing==2.2.0
fastapi==0.115.8
faster-whisper==0.9.0
fastjsonschema==2.21.1
ffmpy==0.5.0
filelock==3.17.0
Flask==3.1.0
flatbuffers==25.1.24
fonttools==4.55.8
fqdn==1.5.1
fsspec==2025.2.0
gradio==3.48.0
gradio_client==0.6.1
h11==0.14.0
httpcore==1.0.7
httpx==0.28.1
huggingface-hub==0.17.3
humanfriendly==10.0
idna==3.10
importlib_metadata==8.6.1
importlib_resources==6.5.2
inflect==7.0.0
ipykernel==6.29.5
ipython==8.18.1
isoduration==20.11.0
itsdangerous==2.2.0
jedi==0.19.2
jieba==0.42.1
Jinja2==3.1.5
jiter==0.8.2
joblib==1.4.2
json5==0.10.0
jsonpointer==3.0.0
jsonschema==4.23.0
jsonschema-specifications==2024.10.1
jupyter-events==0.12.0
jupyter-lsp==2.2.5
jupyter_client==8.6.3
jupyter_core==5.7.2
jupyter_server==2.15.0
jupyter_server_terminals==0.5.3
jupyterlab==4.3.5
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.3
kiwisolver==1.4.7
langid==1.1.6
librosa==0.9.1
llvmlite==0.43.0
MarkupSafe==2.1.5
matplotlib==3.8.4
matplotlib-inline==0.1.7
mistune==3.1.1
more-itertools==10.6.0
mpmath==1.3.0
narwhals==1.25.0
nbclient==0.10.2
nbconvert==7.16.6
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.2.1
notebook==7.3.2
notebook_shim==0.2.4
numba==0.60.0
numpy==1.22.0
nvidia-cublas-cu12==12.4.5.8
nvidia-cuda-cupti-cu12==12.4.127
nvidia-cuda-nvrtc-cu12==12.4.127
nvidia-cuda-runtime-cu12==12.4.127
nvidia-cudnn-cu12==9.1.0.70
nvidia-cufft-cu12==11.2.1.3
nvidia-curand-cu12==10.3.5.147
nvidia-cusolver-cu12==11.6.1.9
nvidia-cusparse-cu12==12.3.1.170
nvidia-cusparselt-cu12==0.6.2
nvidia-nccl-cu12==2.21.5
nvidia-nvjitlink-cu12==12.4.127
nvidia-nvtx-cu12==12.4.127
onnxruntime==1.19.2
openai==1.61.0
openai-whisper==20240930
orjson==3.10.15
overrides==7.7.0
packaging==24.2
pandas==2.0.3
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
pillow==10.4.0
platformdirs==4.3.6
pooch==1.8.2
proces==0.1.7
prometheus_client==0.21.1
prompt_toolkit==3.0.50
protobuf==5.29.3
psutil==6.1.1
ptyprocess==0.7.0
pure_eval==0.2.3
pycparser==2.22
pydantic==2.10.6
pydantic_core==2.27.2
pydub==0.25.1
Pygments==2.19.1
pyparsing==3.2.1
pypinyin==0.50.0
python-dateutil==2.9.0.post0
python-dotenv==1.0.1
python-json-logger==3.2.1
python-multipart==0.0.20
pytz==2025.1
PyYAML==6.0.2
pyzmq==26.2.1
referencing==0.36.2
regex==2024.11.6
requests==2.32.3
resampy==0.4.3
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rpds-py==0.22.3
scikit-learn==1.6.1
scipy==1.11.4
semantic-version==2.10.0
Send2Trash==1.8.3
six==1.17.0
sniffio==1.3.1
soundfile==0.13.1
soupsieve==2.6
stack-data==0.6.3
starlette==0.45.3
sympy==1.13.1
terminado==0.18.1
threadpoolctl==3.5.0
tiktoken==0.8.0
tinycss2==1.4.0
tokenizers==0.14.1
tomli==2.2.1
torch==2.6.0
torchaudio==2.6.0
tornado==6.4.2
tqdm==4.67.1
traitlets==5.14.3
triton==3.2.0
types-python-dateutil==2.9.0.20241206
typing_extensions==4.12.2
tzdata==2025.1
Unidecode==1.3.7
uri-template==1.3.0
urllib3==2.3.0
uvicorn==0.34.0
wavmark==0.0.3
wcwidth==0.2.13
webcolors==24.11.1
webencodings==0.5.1
websocket-client==1.8.0
websockets==11.0.3
Werkzeug==3.1.3
whisper-timestamped==1.14.2
zipp==3.21.0
'''



ORIGINAL_DEPENDENCIES = '''
librosa==0.9.1
faster-whisper==0.9.0
pydub==0.25.1
wavmark==0.0.3
numpy==1.22.0
eng_to_ipa==0.0.2
inflect==7.0.0
unidecode==1.3.7
whisper-timestamped==1.14.2
openai
python-dotenv
pypinyin==0.50.0
cn2an==0.5.22
jieba==0.42.1
gradio==3.48.0
langid==1.1.6
flask
notebook

'''


from .planner_action import IPlannerAction
from dataclasses import dataclass
from ..core import Core

@dataclass
class StopCommand(IPlannerAction):
    instance_id: str

    def apply(self, core: Core):
        op_state = core.operator_states[self.instance_id]
        core.operator_log.decider(op_state.key).event("Stoping")
        op_state.exit_request = True
        op_state.operator_thread.join()
        op_state.controller.stop(self.instance_id)
        del core.operator_states[self.instance_id]
        core.operator_log.decider(op_state.key).event("Stopped")


from .step import IStep, IStepFactory
from .simple_steps import FilterStep, ConstantStep
from .equal_representation_step import ProbabilisticEqualRepresentationStep, AbstractRepresentationStep
from .brainbox_steps import BrainBoxMappingStep, PromptMappingStepFactory


import os
import shutil
from pathlib import Path

from brainbox.framework import OnDemandDockerApi, FileLike, FileIO, RunConfiguration
from .controller import WhisperXController
from .settings import WhisperXSettings
from ...common import check_hf_model_access

class WhisperX(OnDemandDockerApi[WhisperXSettings, WhisperXController]):
    def __init__(self):
        pass

    def _get_dst_folder(self):
        return self.controller.resource_folder()

    def _make_output_dir(self, dirname: FileLike.Type):
        output_dir = Path(self._get_dst_folder() / dirname)
        os.makedirs(output_dir, exist_ok=True)

    def _move_file(self, filename: FileLike.Type):
        path = FileLike.get_path(FileLike(filename, self.cache_folder))
        target_dir = Path(self._get_dst_folder()) / 'files'
        shutil.rmtree(target_dir)
        os.makedirs(target_dir, exist_ok=True)
        shutil.copy(path, target_dir)

    def execute(self, filename: FileLike.Type):
        self._move_file(filename)
        self._make_output_dir('output')

        if 'HF_TOKEN' not in os.environ:
            raise ValueError("""
To download the whisperx models, you need to create a Hugging Face token here https://huggingface.co/settings/tokens

Then, you need to place this token in HF_TOKEN environmental variable
            """)
        token=os.environ['HF_TOKEN']
        check_hf_model_access('pyannote/segmentation-3.0', token)
        check_hf_model_access('pyannote/speaker-diarization-3.1', token)

        configuration = RunConfiguration(
            detach_and_interactive=False,
            command_line_arguments=['--hf_token', token]
        )
        self.controller.run_with_configuration(configuration)

        return FileIO.read_text(self.controller.resource_folder() / self._get_dst_folder() / 'output' / 'output.json')


    Controller = WhisperXController
    Settings = WhisperXSettings



from dataclasses import dataclass
from ..core import DeciderInstanceKey, Core, Job
from .planner_action import IPlannerAction
from sqlalchemy import select
from datetime import datetime
import copy

@dataclass
class FailedJobArgument:
    job: Job

@dataclass
class AssignAction(IPlannerAction):
    job_id: str
    instance_id: str
    key: DeciderInstanceKey

    def apply(self, core: Core):
        with core.new_session() as session:
            job = core.get_job_by_id(session, self.job_id)
            if job.finished:
                return
            arguments = job.arguments
            if job.dependencies is not None:
                requirements = list(job.dependencies.values())
                requirement_to_result = {}
                for element in session.scalars(select(Job).where(Job.id.in_(requirements))):
                    if element.success:
                        requirement_to_result[element.id] = element.result
                    else:
                        session.expunge(element)
                        requirement_to_result[element.id] = FailedJobArgument(element)

                for arg_name, id in job.dependencies.items():
                    if id not in requirement_to_result:
                        core.close_job(
                            session,
                            job,
                            f'Something is wrong: task was ready, but dependency {id} for argument {arg_name} was not found'
                        )
                        return
                    if isinstance(arg_name, str) and len(arg_name) > 0 and arg_name[0] != '*':
                        arguments[arg_name] = requirement_to_result[id]

            try:
                job_clone = copy.copy(job)
                core.operator_states[self.instance_id].jobs_queue.put(job_clone)
                job.assigned = True
                job.assigned_timestamp = datetime.now()
                core.operator_log.task(self.job_id).event('Assigned')
            except:
                core.close_job(session, job, "Assignment failed")

            finally:
                session.commit()


from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase

def run_without_parameters(test_case:TestCase, api: BrainBox.Api):
    """
    ### Call an endpoint

    This will execute the method HelloBrainBox::json on the server side.
    """

    result = api.execute(BrainBox.Task.call(HelloBrainBox).json("Hello"))
    test_case.assertDictEqual(
        {
            'argument': 'Hello',
            'model': 'no_parameter',
            'setting': 'default_setting'
        },
        result
    )

    """
    The BrainBox will run the container for the HelloBrainBox decider.
    This container has a web-server inside, and HelloBrainBox::json 
    connects to this web-server and performs the required operation.
    The result (in this case, `dict`) is then returned to the caller.
    As you see, `argument` equals to the argument the code provides.
    
    BrainBox is smart and tries to minimize the containers' runs.
    If several tasks for same decider have arrived,
    this decider will be set up, and then all these tasks will be executed
    before the tasks for other deciders.
    """

def run_with_parameter(test_case: TestCase, api: BrainBox.Api):
    """
    ### Call an enpoint with a parameter

    Few deciders have parameters: a string argument that will be fed to the container
    on the start, and may modify its behaviour. It's sometimes required
    when you don't build container from scratch, but use a ready container,
    which requires you to load the model on the startup, and in this case,
    `parameter` is this model's name.

    To pass the parameter, simply add it to `call` function:
    """

    result = api.execute(BrainBox.Task.call(HelloBrainBox, "parameter").json("Hello"))
    test_case.assertDictEqual(
        {
            'argument': 'Hello',
            'model': 'parameter',
            'setting': 'default_setting'
        },
        result
    )

    """
    As you can see, the `model` field is now set to `parameter`. 
    
    Most of the deciders do not have the parameter and load models dynamically, so normally you write
    `BrainBoxTask.call(DeciderClass).method(...)`
    
    Please ignore the `setting` field. 
    There is a possibility to adjust settings to the controllers,
    such as the port the internal web-server will be assigned for,
    or the list of models the controllers need to download,
    but so far there was never a necessity to adjust them manually.
    They are only used with the default values.
    """



import shutil
from ....framework import OnDemandDockerApi, ResourcePrerequisite, FileLike, FileIO

from .controller import VideoToImagesController
from .settings import VideoToImagesSettings
from yo_fluq import Query
import json
from .container.settings import AnalysisSettings

class VideoToImages(OnDemandDockerApi[VideoToImagesSettings, VideoToImagesController]):
    def __init__(self):
        pass

    def _monitor(self, s):
        try:
            number = int(s.split('%')[0])
            self.context.logger.report_progress(number/100)
        except:
            pass


    def process(self, analysis_settings: AnalysisSettings|dict):
        if isinstance(analysis_settings, dict):
            analysis_settings = AnalysisSettings(**analysis_settings)

        FileIO.write_json(analysis_settings.__dict__, self.controller.resource_folder()/'settings.json')
        self.run_container(self.controller.get_run_configuration(), self._monitor)

        folder = self.controller.resource_folder('output/images')

        text = FileIO.read_text(self.controller.resource_folder('output')/'frames.jsonl')
        lines = text.split('\n')
        records = []

        for line in lines:
            if line.strip() == '':
                continue
            try:
                record = json.loads(line)
            except:
                print(f"Error parsing line\n`{line}`")
                continue

            fname = self.current_job_id+'.'+record['filename']
            shutil.move(folder/record['filename'], self.cache_folder / fname)
            record['filename'] = fname
            records.append(record)

        return records


    @staticmethod
    def upload_video(file: FileLike.Type):
        return ResourcePrerequisite(
            VideoToImages,
            f'/input/{FileLike.get_name(file,True)}',
            file
        )

    @staticmethod
    def delete_video(filename: str):
        return ResourcePrerequisite(
            VideoToImages,
            '/input/'+filename,
            None
        )

    Controller = VideoToImagesController
    Settings = VideoToImagesSettings
    AnalysisSettings = AnalysisSettings


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder,
    BrainBoxApi, BrainBoxTask, File,
    OnDemandDockerController
)
from .settings import VideoToImagesSettings
from pathlib import Path



class VideoToImagesController(OnDemandDockerController[VideoToImagesSettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_default_settings(self):
        return VideoToImagesSettings()

    def post_install(self):
        if (self.resource_folder('cache')/'huggingface/hub/models--sentence-transformers--clip-ViT-B-32').is_dir():
            return
        self.run_with_configuration(self.get_run_configuration(['--install']))

    def create_api(self):
        from .api import VideoToImages
        return VideoToImages()

    def get_run_configuration(self, arguments: list[str] | None = None):
        if arguments is None:
            arguments = []
        return RunConfiguration(
            None,
            command_line_arguments=arguments,
            detach_and_interactive=False,
            mount_resource_folders={
                'cache' : '/home/app/.cache'
            }
        )


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import VideoToImages

        VideoToImages.upload_video(Path(__file__).parent/'sample.mp4').execute(api)
        result = api.execute(BrainBoxTask.call(VideoToImages).process(VideoToImages.AnalysisSettings(
            'sample.mp4',
            max_produced_frames=5,
        )))
        tc.assertEqual(5, len(result))
        yield TestReport.last_call(api).href('run')


        result = api.execute(BrainBoxTask.call(VideoToImages).process(VideoToImages.AnalysisSettings(
            'sample.mp4',
            max_produced_frames=5,
            add_semantic_comparator=True,
        )))
        tc.assertEqual(5, len(result))
        yield TestReport.last_call(api).href('run-comparator')


        result = api.execute(BrainBoxTask.call(VideoToImages).process(VideoToImages.AnalysisSettings(
            'sample.mp4',
            max_produced_frames=5,
            buffer_by_laplacian_in_ms=500,
        )))
        tc.assertEqual(5, len(result))
        yield TestReport.last_call(api).href('run_bufferer')


DOCKERFILE = f"""
FROM python:3.11.11

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

WORKDIR /home/app

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . /home/app

ENTRYPOINT ["python3", "/home/app/main.py"]
"""

DEPENDENCIES = """
annotated-types==0.7.0
anyio==4.5.2
certifi==2019.11.28
chardet==3.0.4
click==8.1.7
dbus-next==0.2.3
exceptiongroup==1.2.2
fastapi==0.115.3
filelock==3.16.1
fsspec==2024.10.0
h11==0.14.0
huggingface-hub==0.27.0
idna==2.8
Jinja2==3.1.4
joblib==1.4.2
loguru==0.7.2
MarkupSafe==2.1.5
mpmath==1.3.0
networkx==3.1
numpy==1.24.4
nvidia-cublas-cu12==12.1.3.1
nvidia-cuda-cupti-cu12==12.1.105
nvidia-cuda-nvrtc-cu12==12.1.105
nvidia-cuda-runtime-cu12==12.1.105
nvidia-cudnn-cu12==9.1.0.70
nvidia-cufft-cu12==11.0.2.54
nvidia-curand-cu12==10.3.2.106
nvidia-cusolver-cu12==11.4.5.107
nvidia-cusparse-cu12==12.1.0.106
nvidia-nccl-cu12==2.20.5
nvidia-nvjitlink-cu12==12.6.85
nvidia-nvtx-cu12==12.1.105
opencv-contrib-python-headless==4.10.0.82
packaging==24.2
Pillow==10.1.0
pydantic==2.9.2
pydantic_core==2.23.4
PyYAML==6.0.2
regex==2024.11.6
requests==2.22.0
requests-unixsocket==0.2.0
safetensors==0.4.5
scikit-learn==1.3.2
scipy==1.10.1
sentence-transformers==3.2.0
six==1.14.0
sniffio==1.3.1
starlette==0.41.3
sympy==1.13.3
threadpoolctl==3.5.0
tokenizers==0.20.3
torch==2.4.1
tqdm==4.67.1
transformers==4.46.3
triton==3.0.0
typing_extensions==4.12.2
urllib3==1.25.8
uvicorn==0.32.0
"""


from .image_source import IImageSource
from ..executor import IExecutor, Command


class LocalImageSource(IImageSource):
    def __init__(self, image_name: str):
        self.image_name = image_name

    def get_image_name(self):
        return self.image_name

    def get_container_name(self):
        return self.image_name

    def pull(self, executor: IExecutor):
        pass

    def push(self, executor: IExecutor):
        pass









from pathlib import Path
import subprocess
import os
from converter import Converter
import pickle


def concat_with_mpeg(temp_file, output_path, paths):
    control_content_array = [f"file '{p}'" for p in paths]
    control_content = '\n'.join(control_content_array)
    with open(temp_file, 'w') as stream:
        stream.write(control_content)

    args = [
        'ffmpeg',
        '-f',
        'concat',
        '-safe',
        '0',
        '-i',
        str(temp_file),
        '-y',
        str(output_path)
    ]
    try:
        subprocess.check_output(args)
    except subprocess.CalledProcessError as err:
        raise ValueError(f'FFMpeg returned non-zero value. arguments are\n{" ".join(args)}. Output\n{err.output}')


def train(voice: str, converter: Converter):
    source_folder = Path(f'/voices/{voice}/')
    paths = [source_folder/f for f in os.listdir(source_folder)]
    temp_wav = f'/temp/{voice}.wav'
    concat_with_mpeg('files.tmp', temp_wav, paths)
    os.unlink('files.tmp')
    model = converter.convert(temp_wav)
    with open(f'/models/{voice}', 'wb') as stream:
        pickle.dump(model, stream)











from dataclasses import dataclass
from ....framework import ConnectionSettings
from .model import YoloModel

@dataclass
class YoloSettings:
    connection = ConnectionSettings(20302)
    models_to_downloads = (
        YoloModel('arnabdhar/YOLOv8-Face-Detection', 'model.pt'), #Photorealistic faces
        YoloModel('Fuyucchi/yolov8_animeface', 'yolov8x6_animeface.pt') #Anime faces
    )


import os
from typing import *
from foundation_kaia.marshalling import Api, bind_to_api, TestApi
from brainbox.framework import Loc, Locator
from .interface import IControllerService
from .service import ControllerService, ControllerServerSettings
from .server import ControllerServer
from ...common import IDecider, File, FileLike
from ..controller import IController, ControllerRegistry
from pathlib import Path
import requests

@bind_to_api(ControllerService)
class ControllerApi(Api, IControllerService):
    def __init__(self, address: str):
        super().__init__(address)

    def download_resource(self,
                 decider: str|IDecider|IController|type[IDecider]|type[IController],
                 path: str,
                 target_path: Path|str
                 ):
        decider = ControllerRegistry.to_controller_name(decider)
        address = f'http://{self.address}/resources/download/{decider}/{path}'
        response = requests.get(address)
        if response.status_code != 200:
            raise ValueError(f"Couldn't get file content for {address}\n" + response.text)
        content = response.content
        os.makedirs(Path(target_path).parent, exist_ok=True)
        with open(target_path, 'wb') as stream:
            stream.write(content)


    def upload_resource(self,
               decider: str | IDecider | IController | type[IDecider] | type[IController],
               path: str,
               data: FileLike.Type
               ):
        decider = ControllerRegistry.to_controller_name(decider)
        with FileLike(data, None) as stream:
            reply = requests.post(
                f'http://{self.address}/resources/upload/{decider}/{path}',
                files=(
                    ('filename', stream),
                )
            )
        if reply.status_code != 200:
            raise ValueError(reply.text)

    def install_if_not_installed(self, decider: str|IDecider|IController|type[IDecider]|type[IController], with_self_test: bool = True):
        if not self.controller_status(decider).installation_status.installed:
            self.install(decider)
            if with_self_test:
                self.self_test(decider)


    class Test(TestApi['ControllerApi']):
        def __init__(self,
                     services: Iterable[IDecider|IController]|None = None,
                     custom_folder: Path|None = None
                     ):
            if custom_folder is not None:
                loc = Locator(custom_folder)
            else:
                loc = Loc
            registry = ControllerRegistry.discover_or_create(services)
            registry.locator = loc
            settings = ControllerServerSettings(registry)
            super().__init__(ControllerApi, ControllerServer(settings))



from typing import *
from ....framework import BrainBoxTask, BrainBoxCombinedTask, IBrainBoxTask, BrainBoxExtendedTask, CombinedPrerequisite, CacheUploadPrerequisite
from pathlib import Path
import dataclasses



class TaskBuilder:
    @dataclasses.dataclass
    class Record:
        task: IBrainBoxTask
        tags: None | dict[str, Any]

    def __init__(self,
                 records: Optional[Iterable['TaskBuilder.Record']] = None,
                 limit_records_for_test_purposes_at: None | int = None
                 ):
        self.records = [] if records is None else list(records)
        self.limit_records_for_test_purposes_at = limit_records_for_test_purposes_at
        self.uploads: dict[str, Path] = {}

    def _can_add(self):
        if self.limit_records_for_test_purposes_at is None:
            return True
        return len(self.records) < self.limit_records_for_test_purposes_at

    def append(self, task: IBrainBoxTask, tags: None | dict[str, Any]):
        if self._can_add():
            self.records.append(TaskBuilder.Record(task, tags))
        return task

    def to_collector_pack(self, method: str, **kwargs):
        dependencies = {}
        tags = {}
        tasks = []

        for record in self.records:
            tasks.append(record.task)
            if record.tags is not None:
                tags[record.task.get_resulting_id()] = record.tags
                dependencies[record.task.get_resulting_id()] = record.task.get_resulting_id()

        resulting_task = BrainBoxTask(
            decider='Collector',
            decider_method=method,
            dependencies=dependencies,
            arguments=dict(tags=tags, **kwargs)
        )

        result = BrainBoxCombinedTask(resulting_task, tuple(tasks))
        if len(self.uploads) > 0:
            result = BrainBoxExtendedTask(
                result,
                CombinedPrerequisite([
                    CacheUploadPrerequisite(path,filename) for filename, path in self.uploads.items()
                ])
            )
        return result






from typing import Iterable
from unittest import TestCase

from ....framework import (
    TestReport, SmallImageBuilder, IImageBuilder,
    BrainBoxApi, BrainBoxTask,
    OnDemandDockerController
)
from .settings import WhisperXSettings
from pathlib import Path



class WhisperXController(OnDemandDockerController[WhisperXSettings]):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_default_settings(self):
        return WhisperXSettings()

    def create_api(self):
        from .api import WhisperX
        return WhisperX()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import WhisperX

        filename = Path(__file__).parent/'files/sample.mp4'
        api.execute(BrainBoxTask.call(WhisperX).execute(filename))
        yield TestReport.last_call(api).href('run')




DOCKERFILE = f"""
FROM python:3.11.11

# Add NVIDIA APT repo and install cuDNN + CUDA runtime
RUN apt-get update && apt-get install -y curl gnupg && \
    mkdir -p /etc/apt/keyrings && \
    curl -fsSL https://developer.download.nvidia.com/compute/cuda/repos/ubuntu2204/x86_64/3bf863cc.pub \
    | gpg --dearmor -o /etc/apt/keyrings/cuda-archive-keyring.gpg && \
    echo "deb [signed-by=/etc/apt/keyrings/cuda-archive-keyring.gpg] https://developer.download.nvidia.com/compute/cuda/repos/ubuntu2204/x86_64 /" \
    > /etc/apt/sources.list.d/cuda.list && \
    apt-get update && \
    apt-get install -y libcudnn8=8.9.4.* cuda-runtime-12-1 && \
    apt-get clean && rm -rf /var/lib/apt/lists/*
    
{SmallImageBuilder.APT_INSTALL('ffmpeg')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

WORKDIR /home/app

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . .

ENTRYPOINT ["python3", "/home/app/main.py"]
"""

DEPENDENCIES = """
--extra-index-url https://download.pytorch.org/whl/cu121
aiohappyeyeballs==2.6.1
aiohttp==3.12.12
aiosignal==1.3.2
alembic==1.16.1
antlr4-python3-runtime==4.9.3
asteroid-filterbanks==0.4.0
attrs==25.3.0
av==14.4.0
certifi==2025.4.26
cffi==1.17.1
charset-normalizer==3.4.2
click==8.2.1
coloredlogs==15.0.1
colorlog==6.9.0
contourpy==1.3.2
ctranslate2==4.4.0
cycler==0.12.1
docopt==0.6.2
einops==0.8.1
faster-whisper==1.1.1
filelock==3.18.0
flatbuffers==25.2.10
fonttools==4.58.2
frozenlist==1.7.0
fsspec==2025.5.1
greenlet==3.2.3
hf-xet==1.1.2
huggingface-hub==0.33.0
humanfriendly==10.0
HyperPyYAML==1.2.2
idna==3.10
Jinja2==3.1.6
joblib==1.5.1
julius==0.2.7
kiwisolver==1.4.8
lightning==2.5.1.post0
lightning-fabric==2.0.4
lightning-utilities==0.14.3
Mako==1.3.10
markdown-it-py==3.0.0
MarkupSafe==3.0.2
matplotlib==3.10.3
mdurl==0.1.2
mpmath==1.3.0
multidict==6.4.4
networkx==3.5
nltk==3.9.1
numpy==2.3.0
nvidia-cublas-cu12==12.6.4.1
nvidia-cuda-cupti-cu12==12.6.80
nvidia-cuda-nvrtc-cu12==12.6.77
nvidia-cuda-runtime-cu12==12.6.77
nvidia-cudnn-cu12==9.5.1.17
nvidia-cufft-cu12==11.3.0.4
nvidia-cufile-cu12==1.11.1.6
nvidia-curand-cu12==10.3.7.77
nvidia-cusolver-cu12==11.7.1.2
nvidia-cusparse-cu12==12.5.4.2
nvidia-cusparselt-cu12==0.6.3
nvidia-nccl-cu12==2.26.2
nvidia-nvjitlink-cu12==12.6.85
nvidia-nvtx-cu12==12.6.77
omegaconf==2.3.0
onnxruntime==1.22.0
optuna==4.3.0
packaging==24.2
pandas==2.3.0
pillow==11.2.1
primePy==1.3
propcache==0.3.2
protobuf==6.31.1
pyannote.audio==3.3.2
pyannote.core==5.0.0
pyannote.database==5.1.3
pyannote.metrics==3.2.1
pyannote.pipeline==3.0.1
pycparser==2.22
Pygments==2.19.1
pyparsing==3.2.3
python-dateutil==2.9.0.post0
pytorch-lightning==2.5.1.post0
pytorch-metric-learning==2.8.1
pytz==2025.2
PyYAML==6.0.2
regex==2024.11.6
requests==2.32.4
rich==14.0.0
ruamel.yaml==0.18.14
ruamel.yaml.clib==0.2.12
safetensors==0.5.3
scikit-learn==1.7.0
scipy==1.15.3
semver==3.0.4
sentencepiece==0.2.0
shellingham==1.5.4
six==1.17.0
sortedcontainers==2.4.0
soundfile==0.13.1
speechbrain==1.0.3
SQLAlchemy==2.0.41
sympy==1.14.0
tabulate==0.9.0
tensorboardX==2.6.4
threadpoolctl==3.6.0
tokenizers==0.21.1
torch==2.7.1
torch-audiomentations==0.12.0
torch_pitch_shift==1.2.5
torchaudio==2.7.1
torchmetrics==1.7.2
tqdm==4.67.1
transformers==4.52.4
triton==3.3.1
typer==0.16.0
typing_extensions==4.14.0
tzdata==2025.2
urllib3==2.4.0
whisperx==3.3.4
yarl==1.20.1
"""


ORIGINAL_DEPENDENCIES = """
whisperx==3.3.4
lightning_fabric==2.0.4
hf_xet==1.1.2
"""


from typing import *
from .task import IBrainBoxTask
from abc import ABC, abstractmethod
from ...job_processing import Job, OperatorLogItem
from typing import *
from dataclasses import dataclass

@dataclass
class _ArrayPostprocessor:
    index: int
    api: Any
    postprocessor: Any = None

    def __call__(self, array):
        if self.index is None:
            return None
        result = array[self.index]
        if self.postprocessor is not None:
            result = self.postprocessor(result, self.api)
        return result


class IBrainboxService(ABC):
    @abstractmethod
    def base_add(self, jobs: list[dict]):
        pass

    @abstractmethod
    def base_join(self, ids: list[str]) -> list:
        pass

    @abstractmethod
    def result(self, id: str):
        pass

    @abstractmethod
    def job(self, id: str) -> Job:
        pass

    @abstractmethod
    def summary(self, ids: list[str]|None = None, batch_id: str|None = None) -> list[dict]:
        pass

    @abstractmethod
    def cancel(self, job_id: str|None = None, batch_id: str|None = None):
        pass

    @abstractmethod
    def get_operator_log(self, entries_count: int = 100) -> list[OperatorLogItem]:
        pass

    @abstractmethod
    def shutdown(self):
        pass

    def add(self, task: Union[IBrainBoxTask, Iterable[IBrainBoxTask]]) -> str|list[str]:
        if isinstance(task, IBrainBoxTask):
            single = True
            task.before_add(self)
            task = [task]
        else:
            single = False
            try:
                task = list(task)
            except:
                raise ValueError(f"task should be IBrainBoxTask or iterable, but was: {task}")
            for index, t in enumerate(task):
                if not isinstance(t, IBrainBoxTask):
                    raise ValueError(f"Expected task at position #{index}, but was {t}")
                t.before_add(self)

        job_dicts = []
        resulting_ids = []
        for t in task:
            jobs = t.create_jobs()
            batch_id = t.get_resulting_id()
            resulting_ids.append(batch_id)
            for j in jobs:
                j.batch = batch_id
                d = j.__dict__
                del d['_sa_instance_state']
                job_dicts.append(d)

        self.base_add(job_dicts)
        if single:
            return resulting_ids[0]
        else:
            return resulting_ids



    def join(self, task: Union[IBrainBoxTask, str, Iterable[Union[IBrainBoxTask, str]]]):
        if isinstance(task, str) or isinstance(task, IBrainBoxTask):
            task = [task]
            not_list = True
        else:
            try:
                task = list(task)
            except:
                raise ValueError(f"Task is expected to be str (job id), IBrainBoxTask or Iterable, but was {task}")
            not_list = False

        ids = []
        postprocessors = []
        index = 0
        for true_index, t in enumerate(task):
            if isinstance(t, str):
                ids.append(t)
                postprocessors.append(_ArrayPostprocessor(index, self))
                index += 1
            elif isinstance(t, IBrainBoxTask):
                id = t.get_resulting_id()
                if id is None:
                    postprocessors.append(_ArrayPostprocessor(None, self))
                else:
                    ids.append(t.get_resulting_id())
                    postprocessors.append(_ArrayPostprocessor(index, self, t.postprocess_result))
                    index+=1
            else:
                raise ValueError(f"Error at index {true_index}: expected str (job id) or IBrainBoxTask, but was {t}")

        result = self.base_join(ids)
        result = [postproc(result) for postproc in postprocessors]

        if not_list:
            return result[0]
        return result

    def execute(self, task: Union[IBrainBoxTask, Iterable[IBrainBoxTask]]):
        self.add(task)
        return self.join(task)





from dataclasses import dataclass, field
from .decider_instance_key import DeciderInstanceKey
from datetime import datetime
from queue import Queue
from .job import Job
from .operator_message import OperatorMessage
from .operator_log import OperatorLog
from ...controllers.controller import IController
from ...common import IDecider
from threading import Thread

@dataclass
class OperatorStateForPlanner:
    instance_id: str
    key: DeciderInstanceKey
    busy: bool
    not_busy_since: datetime|None


@dataclass
class OperatorState:
    key: DeciderInstanceKey
    controller: IController
    controller_run_instance_id: str
    api: IDecider
    logger: OperatorLog = field(default_factory=OperatorLog)
    operator_thread: Thread|None = None
    exit_request: bool = False
    busy: bool = False
    not_busy_since: datetime|None = None
    jobs_queue: Queue[Job] = field(default_factory=Queue)
    results_queue: Queue[OperatorMessage] = field(default_factory=Queue)








if __name__ == '__main__':
    with open('/resources/input','r') as input:
        data = input.read()
    for i in range(100):
        print(f'{i}%')
    with open('/resources/output', 'w') as output:
        output.write('output '+data)


import os
from pathlib import Path
import requests
from ....framework import DockerWebServiceApi, FileLike, File, FileIO
from .settings import ResembleEnhanceSettings
from .controller import ResembleEnhanceController
from uuid import uuid4

class ResembleEnhance(DockerWebServiceApi[ResembleEnhanceSettings, ResembleEnhanceController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)

    def _process(self, file: FileLike.Type):
        uploads = self.controller.resource_folder('uploads')
        os.makedirs(uploads, exist_ok=True)
        for old_file in os.listdir(uploads):
            os.unlink(uploads/old_file)
        fname = str(uuid4())+'.wav'
        with FileLike(file, self.cache_folder) as stream:
            data = stream.read()
            with open(uploads/fname, 'wb') as out_stream:
                out_stream.write(data)
        reply = requests.post(f'http://{self.address}/enhance', json=dict(filename=fname))
        if reply.status_code != 200:
            raise ValueError(f"Decider returned code {reply.status_code}\n{reply.text}")
        js = reply.json()
        return js

    def process(self, file: FileLike.Type):
        return [
            File(name, FileIO.read_bytes(self.controller.resource_folder('outputs')/name))
            for name in self._process(file)
        ]

    def enhance(self, file: FileLike.Type):
        name = self._process(file)[-1]
        return File(name, FileIO.read_bytes(self.controller.resource_folder('outputs')/name))





    Settings = ResembleEnhanceSettings
    Controller = ResembleEnhanceController


from ...job_processing import Job
from .task import IBrainBoxTask
from abc import ABC, abstractmethod

class IOneBrainBoxTaskFactory(IBrainBoxTask, ABC):
    @abstractmethod
    def to_task(self):
        pass

    def create_jobs(self) -> list[Job]:
        return self.to_task().create_jobs()

    def get_resulting_id(self) -> str:
        return self.to_task().get_resulting_id()

    def before_add(self, api):
        return self.to_task().before_add(api)

    def postprocess_result(self, result, api):
        return self.to_task().postprocess_result(result, api)


from typing import Generic, TypeVar, Optional,Callable
from .cache_manager import ICacheManager
from abc import ABC, abstractmethod
import hashlib

TExternalKey = TypeVar('TExternalKey')
TInternalKey = TypeVar('TInternalKey')

class Accessor(Generic[TExternalKey, TInternalKey], ABC):
    def __init__(self, cache: ICacheManager[TInternalKey]):
        self.cache = cache

    def _hash_str(self, s: str):
        return hashlib.sha256(str(s).encode()).hexdigest()

    def _translate_external_key(self, key: TExternalKey) -> TInternalKey:
        return self._hash_str(str(key))

    @abstractmethod
    def _external_get(self, key: TExternalKey):
        pass

    def _postprocess(self, value):
        return value


    def get(self, key: TExternalKey):
        internal_key = self._translate_external_key(key)
        if not self.cache.contains(internal_key):
            value = self._external_get(key)
            self.cache.store(internal_key, value)
            if not self.cache.contains(internal_key):
                raise ValueError(f"Value for {key}/{internal_key} was not stored in cache")
        return self._postprocess(self.cache.unsafe_get(internal_key))










import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

import pytorch_lightning as pl
import torch
from torch import autocast
from torch.nn import functional as F
from torch.utils.data import DataLoader, Dataset, random_split

from .commons import slice_segments
from .dataset import Batch, PiperDataset, UtteranceCollate
from .losses import discriminator_loss, feature_loss, generator_loss, kl_loss
from .mel_processing import mel_spectrogram_torch, spec_to_mel_torch
from .models import MultiPeriodDiscriminator, SynthesizerTrn

_LOGGER = logging.getLogger("vits.lightning")


class VitsModel(pl.LightningModule):
    def __init__(
        self,
        num_symbols: int,
        num_speakers: int,
        # audio
        resblock="2",
        resblock_kernel_sizes=(3, 5, 7),
        resblock_dilation_sizes=(
            (1, 2),
            (2, 6),
            (3, 12),
        ),
        upsample_rates=(8, 8, 4),
        upsample_initial_channel=256,
        upsample_kernel_sizes=(16, 16, 8),
        # mel
        filter_length: int = 1024,
        hop_length: int = 256,
        win_length: int = 1024,
        mel_channels: int = 80,
        sample_rate: int = 22050,
        sample_bytes: int = 2,
        channels: int = 1,
        mel_fmin: float = 0.0,
        mel_fmax: Optional[float] = None,
        # model
        inter_channels: int = 192,
        hidden_channels: int = 192,
        filter_channels: int = 768,
        n_heads: int = 2,
        n_layers: int = 6,
        kernel_size: int = 3,
        p_dropout: float = 0.1,
        n_layers_q: int = 3,
        use_spectral_norm: bool = False,
        gin_channels: int = 0,
        use_sdp: bool = True,
        segment_size: int = 8192,
        # training
        dataset: Optional[List[Union[str, Path]]] = None,
        learning_rate: float = 2e-4,
        betas: Tuple[float, float] = (0.8, 0.99),
        eps: float = 1e-9,
        batch_size: int = 1,
        lr_decay: float = 0.999875,
        init_lr_ratio: float = 1.0,
        warmup_epochs: int = 0,
        c_mel: int = 45,
        c_kl: float = 1.0,
        grad_clip: Optional[float] = None,
        num_workers: int = 1,
        seed: int = 1234,
        num_test_examples: int = 5,
        validation_split: float = 0.1,
        max_phoneme_ids: Optional[int] = None,
        **kwargs,
    ):
        super().__init__()
        self.save_hyperparameters()

        if (self.hparams.num_speakers > 1) and (self.hparams.gin_channels <= 0):
            # Default gin_channels for multi-speaker model
            self.hparams.gin_channels = 512

        # Set up models
        self.model_g = SynthesizerTrn(
            n_vocab=self.hparams.num_symbols,
            spec_channels=self.hparams.filter_length // 2 + 1,
            segment_size=self.hparams.segment_size // self.hparams.hop_length,
            inter_channels=self.hparams.inter_channels,
            hidden_channels=self.hparams.hidden_channels,
            filter_channels=self.hparams.filter_channels,
            n_heads=self.hparams.n_heads,
            n_layers=self.hparams.n_layers,
            kernel_size=self.hparams.kernel_size,
            p_dropout=self.hparams.p_dropout,
            resblock=self.hparams.resblock,
            resblock_kernel_sizes=self.hparams.resblock_kernel_sizes,
            resblock_dilation_sizes=self.hparams.resblock_dilation_sizes,
            upsample_rates=self.hparams.upsample_rates,
            upsample_initial_channel=self.hparams.upsample_initial_channel,
            upsample_kernel_sizes=self.hparams.upsample_kernel_sizes,
            n_speakers=self.hparams.num_speakers,
            gin_channels=self.hparams.gin_channels,
            use_sdp=self.hparams.use_sdp,
        )
        self.model_d = MultiPeriodDiscriminator(
            use_spectral_norm=self.hparams.use_spectral_norm
        )

        # Dataset splits
        self._train_dataset: Optional[Dataset] = None
        self._val_dataset: Optional[Dataset] = None
        self._test_dataset: Optional[Dataset] = None
        self._load_datasets(validation_split, num_test_examples, max_phoneme_ids)

        # State kept between training optimizers
        self._y = None
        self._y_hat = None

    def _load_datasets(
        self,
        validation_split: float,
        num_test_examples: int,
        max_phoneme_ids: Optional[int] = None,
    ):
        if self.hparams.dataset is None:
            _LOGGER.debug("No dataset to load")
            return

        full_dataset = PiperDataset(
            self.hparams.dataset, max_phoneme_ids=max_phoneme_ids
        )
        valid_set_size = int(len(full_dataset) * validation_split)
        train_set_size = len(full_dataset) - valid_set_size - num_test_examples

        self._train_dataset, self._test_dataset, self._val_dataset = random_split(
            full_dataset, [train_set_size, num_test_examples, valid_set_size]
        )

    def forward(self, text, text_lengths, scales, sid=None):
        noise_scale = scales[0]
        length_scale = scales[1]
        noise_scale_w = scales[2]
        audio, *_ = self.model_g.infer(
            text,
            text_lengths,
            noise_scale=noise_scale,
            length_scale=length_scale,
            noise_scale_w=noise_scale_w,
            sid=sid,
        )

        return audio

    def train_dataloader(self):
        return DataLoader(
            self._train_dataset,
            collate_fn=UtteranceCollate(
                is_multispeaker=self.hparams.num_speakers > 1,
                segment_size=self.hparams.segment_size,
            ),
            num_workers=self.hparams.num_workers,
            batch_size=self.hparams.batch_size,
        )

    def val_dataloader(self):
        return DataLoader(
            self._val_dataset,
            collate_fn=UtteranceCollate(
                is_multispeaker=self.hparams.num_speakers > 1,
                segment_size=self.hparams.segment_size,
            ),
            num_workers=self.hparams.num_workers,
            batch_size=self.hparams.batch_size,
        )

    def test_dataloader(self):
        return DataLoader(
            self._test_dataset,
            collate_fn=UtteranceCollate(
                is_multispeaker=self.hparams.num_speakers > 1,
                segment_size=self.hparams.segment_size,
            ),
            num_workers=self.hparams.num_workers,
            batch_size=self.hparams.batch_size,
        )

    def training_step(self, batch: Batch, batch_idx: int, optimizer_idx: int):
        if optimizer_idx == 0:
            return self.training_step_g(batch)

        if optimizer_idx == 1:
            return self.training_step_d(batch)

    def training_step_g(self, batch: Batch):
        x, x_lengths, y, _, spec, spec_lengths, speaker_ids = (
            batch.phoneme_ids,
            batch.phoneme_lengths,
            batch.audios,
            batch.audio_lengths,
            batch.spectrograms,
            batch.spectrogram_lengths,
            batch.speaker_ids if batch.speaker_ids is not None else None,
        )
        (
            y_hat,
            l_length,
            _attn,
            ids_slice,
            _x_mask,
            z_mask,
            (_z, z_p, m_p, logs_p, _m_q, logs_q),
        ) = self.model_g(x, x_lengths, spec, spec_lengths, speaker_ids)
        self._y_hat = y_hat

        mel = spec_to_mel_torch(
            spec,
            self.hparams.filter_length,
            self.hparams.mel_channels,
            self.hparams.sample_rate,
            self.hparams.mel_fmin,
            self.hparams.mel_fmax,
        )
        y_mel = slice_segments(
            mel,
            ids_slice,
            self.hparams.segment_size // self.hparams.hop_length,
        )
        y_hat_mel = mel_spectrogram_torch(
            y_hat.squeeze(1).to("cpu"),
            self.hparams.filter_length,
            self.hparams.mel_channels,
            self.hparams.sample_rate,
            self.hparams.hop_length,
            self.hparams.win_length,
            self.hparams.mel_fmin,
            self.hparams.mel_fmax,
        )
        y_hat_mel = y_hat_mel.to("cuda")
        y = slice_segments(
            y,
            ids_slice * self.hparams.hop_length,
            self.hparams.segment_size,
        )  # slice

        # Save for training_step_d
        self._y = y

        _y_d_hat_r, y_d_hat_g, fmap_r, fmap_g = self.model_d(y, y_hat)

        with autocast(self.device.type, enabled=False):
            # Generator loss
            loss_dur = torch.sum(l_length.float())
            loss_mel = F.l1_loss(y_mel, y_hat_mel) * self.hparams.c_mel
            loss_kl = kl_loss(z_p, logs_q, m_p, logs_p, z_mask) * self.hparams.c_kl

            loss_fm = feature_loss(fmap_r, fmap_g)
            loss_gen, _losses_gen = generator_loss(y_d_hat_g)
            loss_gen_all = loss_gen + loss_fm + loss_mel + loss_dur + loss_kl

            self.log("loss_gen_all", loss_gen_all)

            return loss_gen_all

    def training_step_d(self, batch: Batch):
        # From training_step_g
        y = self._y
        y_hat = self._y_hat
        y_d_hat_r, y_d_hat_g, _, _ = self.model_d(y, y_hat.detach())

        with autocast(self.device.type, enabled=False):
            # Discriminator
            loss_disc, _losses_disc_r, _losses_disc_g = discriminator_loss(
                y_d_hat_r, y_d_hat_g
            )
            loss_disc_all = loss_disc

            self.log("loss_disc_all", loss_disc_all)

            return loss_disc_all

    def validation_step(self, batch: Batch, batch_idx: int):
        val_loss = self.training_step_g(batch) + self.training_step_d(batch)
        self.log("val_loss", val_loss)

        # Generate audio examples
        for utt_idx, test_utt in enumerate(self._test_dataset):
            text = test_utt.phoneme_ids.unsqueeze(0).to(self.device)
            text_lengths = torch.LongTensor([len(test_utt.phoneme_ids)]).to(self.device)
            scales = [0.667, 1.0, 0.8]
            sid = (
                test_utt.speaker_id.to(self.device)
                if test_utt.speaker_id is not None
                else None
            )
            test_audio = self(text, text_lengths, scales, sid=sid).detach()

            # Scale to make louder in [-1, 1]
            test_audio = test_audio * (1.0 / max(0.01, abs(test_audio.max())))

            tag = test_utt.text or str(utt_idx)
            self.logger.experiment.add_audio(
                tag, test_audio, sample_rate=self.hparams.sample_rate
            )

        return val_loss

    def configure_optimizers(self):
        optimizers = [
            torch.optim.AdamW(
                self.model_g.parameters(),
                lr=self.hparams.learning_rate,
                betas=self.hparams.betas,
                eps=self.hparams.eps,
            ),
            torch.optim.AdamW(
                self.model_d.parameters(),
                lr=self.hparams.learning_rate,
                betas=self.hparams.betas,
                eps=self.hparams.eps,
            ),
        ]
        schedulers = [
            torch.optim.lr_scheduler.ExponentialLR(
                optimizers[0], gamma=self.hparams.lr_decay
            ),
            torch.optim.lr_scheduler.ExponentialLR(
                optimizers[1], gamma=self.hparams.lr_decay
            ),
        ]

        return optimizers, schedulers

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = parent_parser.add_argument_group("VitsModel")
        parser.add_argument("--batch-size", type=int, required=True)
        parser.add_argument("--validation-split", type=float, default=0.1)
        parser.add_argument("--num-test-examples", type=int, default=5)
        parser.add_argument(
            "--max-phoneme-ids",
            type=int,
            help="Exclude utterances with phoneme id lists longer than this",
        )
        #
        parser.add_argument("--hidden-channels", type=int, default=192)
        parser.add_argument("--inter-channels", type=int, default=192)
        parser.add_argument("--filter-channels", type=int, default=768)
        parser.add_argument("--n-layers", type=int, default=6)
        parser.add_argument("--n-heads", type=int, default=2)
        #
        return parent_parser


from dataclasses import dataclass, field
from .test_report_item import TestReportItem

@dataclass
class TestReportSection:
    caption: str = None
    comment: str|None = None
    items: list[TestReportItem] = field(default_factory=list)





from abc import ABC, abstractmethod

class IStep(ABC):
    @abstractmethod
    def process(self, history, current):
        pass

    def shorten(self, data):
        return data

    def summarize(self, data) -> str|None:
        return None

    def get_name(self) -> str|None:
        if not hasattr(self,'_name'):
            return None
        return self._name

    def with_name(self, name: str) -> 'IStep':
        self._name = name
        return self


class IStepFactory(ABC):
    @abstractmethod
    def create_step(self) -> IStep:
        pass



from brainbox import BrainBoxApi, IBrainBoxTask
from brainbox.framework import OnDemandDockerApi, FileLike, FileIO, File, RunConfiguration, IPrerequisite
from .controller import PiperTrainingController
from .settings import PiperTrainingSettings
from .container.settings import TrainingSettings
from .model import PiperTrainingModel
from pathlib import Path
import zipfile
import os
import shutil
import re

class Monitor:
    def __init__(self, api: 'PiperTraining', expected_epochs, folder_name: str):
        self.api = api
        self.expected_epochs = expected_epochs
        self.first_epoch = None
        self.folder_name = folder_name

    def __call__(self, s:str):
        print(s)
        epochs = re.findall(self.folder_name+'/epoch=(\d+)-step', s)
        if len(epochs) > 0:
            try:
                ep = int(epochs[-1])
                if self.first_epoch is None:
                    self.first_epoch = ep
                    return
                percentage = (ep-self.first_epoch)/(self.expected_epochs-self.first_epoch)
                self.api.context.logger.report_progress(percentage)
            except:
                pass



class PiperTraining(OnDemandDockerApi[PiperTrainingSettings, PiperTrainingController]):
    def _unzip_dataset(self, name: str, path: Path, enable_recreate):
        csv = []
        trainings_folder = self.controller.resource_folder('trainings')
        dataset_folder = trainings_folder/name/'source'
        if dataset_folder.is_dir() and not enable_recreate:
            raise ValueError(f"Folder for training {name} exists")
        shutil.rmtree(dataset_folder, ignore_errors=True)
        os.makedirs(dataset_folder/'wav')
        with zipfile.ZipFile(path, 'r') as file:
            for name in file.namelist():
                character, filename = name.split('/')
                if not filename.endswith('.wav'):
                    continue
                data = file.read(name)
                output_path = dataset_folder / 'wav' / filename
                with open(output_path, 'wb') as output_file:
                    output_file.write(data)
                text = file.read(name.replace('.wav', '.txt')).decode('utf-8')
                csv.append(dict(character=character, id=filename, text=text))
        single_speaker = len(set(c['character'] for c in csv)) == 1
        with open(dataset_folder / 'metadata.csv', 'w') as file:
            for record in csv:
                if single_speaker:
                    file.write(f'{record["id"]}|{record["text"]}\n')
                else:
                    file.write(f'{record["id"]}|{record["character"]}|{record["text"]}\n')



    def train(self, name: str, dataset: FileLike.Type, settings: TrainingSettings|dict|None = None, reset_if_existing: bool = False):
        if settings is None:
            settings = TrainingSettings()
        elif isinstance(settings, dict):
            settings = TrainingSettings(**settings)

        path = FileLike(dataset, self.cache_folder).get_path()
        if not settings.continue_existing:
            self._unzip_dataset(name, path, reset_if_existing)
        FileIO.write_json(settings.__dict__, self.controller.resource_folder('trainings', name)/'settings.json')

        configuration = RunConfiguration(
            detach_and_interactive=False,
            mount_top_resource_folder=True,
            command_line_arguments=['--dataset',name]
        )

        max_epochs = settings.max_epochs
        monitor = Monitor(self, max_epochs, 'checkpoints')
        self.controller.run_with_configuration(configuration, monitor)


    def export(self, name: str, max_epochs: int, cleanup: bool = False):
        result = []
        configuration = RunConfiguration(
            detach_and_interactive=False,
            mount_top_resource_folder=True,
            command_line_arguments=['--dataset', name, '--export']
        )
        monitor = Monitor(self, max_epochs, 'result')
        self.controller.run_with_configuration(configuration, monitor)
        result_folder = self.controller.resource_folder('trainings',name,'result')
        for file in os.listdir(result_folder):
            dst_filename = self.current_job_id+'_'+file
            shutil.copy(
                result_folder/file,
                self.cache_folder/dst_filename,
            )
            if file.endswith('.onnx'):
                result.append(dict(onnx=dst_filename, json=dst_filename+'.json'))

        if cleanup:
            shutil.rmtree(self.controller.resource_folder('trainings', name))

        return result

    @staticmethod
    def create_training_pack(
            dataset: FileLike.Type,
            settings: TrainingSettings|dict|None = None,
            custom_name: str|None = None,
            clean_up_afterwards: bool = False,
            reset_if_existing: bool = False
    ):
        name = custom_name if custom_name is not None else FileLike.get_name(dataset, True).split('.')[0]

        from brainbox import BrainBox
        training_task: BrainBox.Task = BrainBox.Task.call(PiperTraining).train(name, dataset, settings, reset_if_existing).to_task()
        id = training_task.id
        training_task.batch = id

        export_task: BrainBox.Task = BrainBox.Task.call(PiperTraining).export(name, settings.max_epochs, clean_up_afterwards).to_task()
        export_task.dependencies['*training'] = id
        export_task.batch = id

        return BrainBox.CombinedTask(
            export_task,
            (training_task,)
        )







    Controller = PiperTrainingController
    Settings = PiperTrainingSettings
    TrainingSettings = TrainingSettings
    Model = PiperTrainingModel



import os
from unittest import TestCase
from ...framework import Loc, FileIO, LocalExecutor
from pathlib import Path

def check_if_its_sound(content, tc: TestCase):
    import soundfile as sf
    with Loc.create_test_file() as tmp_filename:
        FileIO.write_bytes(content, tmp_filename)
        f = sf.SoundFile(tmp_filename)
        duration = f.frames / f.samplerate
        tc.assertGreater(duration, 1)
    f.close()


VOICEOVER_TEXT = 'The quick brown fox jumps over the lazy dog'


def download_file(url: str, file: Path, keep_current_if_exists: bool = True):
    import requests
    from tqdm import tqdm

    file = Path(file)
    if file.is_file():
        if keep_current_if_exists:
            return
        else:
            os.unlink(file)

    response = requests.head(url)
    file_size = int(response.headers.get('Content-Length', 0))  # Get the file size in bytes

    if file_size == 0:
        print("Unable to determine file size. Proceeding without progress bar.")

    # Make the GET request for the file content
    response = requests.get(url, stream=True)
    response.raise_for_status()

    os.makedirs(file.parent, exist_ok=True)

    # Write the file with a progress bar
    with open(file, 'wb') as file, tqdm(
            total=file_size,
            unit='B',
            unit_scale=True,
            desc=str(file)
    ) as progress_bar:
        for chunk in response.iter_content(chunk_size=8192):  # Read in 8 KB chunks
            file.write(chunk)
            progress_bar.update(len(chunk))



from huggingface_hub import HfApi, hf_hub_url
import requests

def check_hf_model_access(repo_id: str, token: str):
    api = HfApi()
    url = hf_hub_url(repo_id=repo_id, filename="config.json")
    headers = {"Authorization": f"Bearer {token}"}
    response = requests.head(url, headers=headers)
    if response.status_code == 403:
        raise ValueError(f"You do not have the access to {repo_id}. Visit https://huggingface.co/{repo_id} and request access")



import json
from typing import Iterable
from unittest import TestCase

from ....framework import (
    File, RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder,
    DockerWebServiceController, BrainBoxApi, BrainBoxTask, IModelDownloadingController, DownloadableModel
)
from .settings import WhisperSettings, WhisperModel
from pathlib import Path


class WhisperController(DockerWebServiceController[WhisperSettings], IModelDownloadingController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return WhisperModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            None,
            publish_ports={self.connection_settings.port:8084},
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return WhisperSettings()

    def create_api(self):
        from .api import Whisper
        return Whisper()

    def run_notebook(self):
        self.run_with_configuration(self.get_service_run_configuration(None).as_notebook_service())

    def post_install(self):
        self.download_models(self.settings.models_to_download)


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Whisper

        file = File.read(Path(__file__).parent / 'files/test_voice.wav')

        first_time = True
        for model in self.settings.models_to_download:
            result = api.execute(BrainBoxTask.call(Whisper).transcribe_json(file, model.name))
            yield TestReport.last_call(api).with_comment("Speech recognition with Whisper, full output")
            tc.assertEqual(
                'One little spark and before you know it, the whole world is burning.',
                result['text'].strip()
            )

            result = api.execute(BrainBoxTask.call(Whisper).transcribe(file, model.name))
            if first_time:
                yield TestReport.last_call(api).href('recognition').with_comment("Speech recognition with Whisper, text-only output")
            tc.assertEqual(
                'One little spark and before you know it, the whole world is burning.',
                result
            )



DOCKERFILE=f'''
FROM python:3.9.21

{SmallImageBuilder.APT_INSTALL('ffmpeg')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . /home/app/

ENTRYPOINT ["python3","/home/app/main.py"]

'''



DEPENDENCIES='''
anyio==4.3.0
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==2.4.1
async-lru==2.0.4
attrs==23.2.0
Babel==2.14.0
beautifulsoup4==4.12.3
bleach==6.1.0
blinker==1.7.0
certifi==2024.2.2
cffi==1.16.0
charset-normalizer==3.3.2
click==8.1.7
colorama==0.4.6
comm==0.2.2
debugpy==1.8.1
decorator==5.1.1
defusedxml==0.7.1
exceptiongroup==1.2.0
executing==2.0.1
fastjsonschema==2.19.1
filelock==3.13.4
Flask==3.0.3
fqdn==1.5.1
fsspec==2024.3.1
h11==0.14.0
httpcore==1.0.5
httpx==0.27.0
idna==3.7
importlib_metadata==7.1.0
ipykernel==6.29.4
ipython==8.18.1
isoduration==20.11.0
itsdangerous==2.1.2
jedi==0.19.1
Jinja2==3.1.3
json5==0.9.25
jsonpointer==2.4
jsonschema==4.21.1
jsonschema-specifications==2023.12.1
jupyter-events==0.10.0
jupyter-lsp==2.2.5
jupyter_client==8.6.1
jupyter_core==5.7.2
jupyter_server==2.14.0
jupyter_server_terminals==0.5.3
jupyterlab==4.1.6
jupyterlab_pygments==0.3.0
jupyterlab_server==2.26.0
llvmlite==0.42.0
MarkupSafe==2.1.5
matplotlib-inline==0.1.6
mistune==3.0.2
more-itertools==10.2.0
mpmath==1.3.0
nbclient==0.10.0
nbconvert==7.16.3
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.2.1
notebook==7.1.2
notebook_shim==0.2.4
numba==0.59.1
numpy==1.26.4
openai-whisper==20231117
overrides==7.7.0
packaging==24.0
pandocfilters==1.5.1
parso==0.8.4
platformdirs==4.2.0
prometheus_client==0.20.0
prompt-toolkit==3.0.43
psutil==5.9.8
pure-eval==0.2.2
pycparser==2.22
Pygments==2.17.2
python-dateutil==2.9.0.post0
python-json-logger==2.0.7
PyYAML==6.0.1
pyzmq==26.0.0
referencing==0.34.0
regex==2023.12.25
requests==2.31.0
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rpds-py==0.18.0
Send2Trash==1.8.3
six==1.16.0
sniffio==1.3.1
soupsieve==2.5
stack-data==0.6.3
sympy==1.12
terminado==0.18.1
tiktoken==0.6.0
tinycss2==1.2.1
tomli==2.0.1
torch==2.2.2
tornado==6.4
tqdm==4.66.2
traitlets==5.14.2
types-python-dateutil==2.9.0.20240316
typing_extensions==4.11.0
uri-template==1.3.0
urllib3==2.2.1
wcwidth==0.2.13
webcolors==1.13
webencodings==0.5.1
websocket-client==1.7.0
Werkzeug==3.0.2
zipp==3.18.1
'''


from dataclasses import dataclass
from .postprocessors import IPostprocessor
from .prerequisites import IPrerequisite
from ..app import IBrainBoxTask
from ...job_processing import Job


@dataclass
class BrainBoxExtendedTask(IBrainBoxTask):
    task: IBrainBoxTask
    prerequisite: IPrerequisite|None = None
    postprocessor: IPostprocessor|None = None

    def __post_init__(self):
        if not isinstance(self.task, IBrainBoxTask):
            raise TypeError("task must be a BrainBoxTask")

    def create_jobs(self) -> list[Job]:
        return self.task.create_jobs()

    def before_add(self, api):
        if self.prerequisite is not None:
            self.prerequisite.execute(api)

    def postprocess_result(self, result, api):
        if self.postprocessor is not None:
            return self.postprocessor.postprocess(result, api)
        return result

    def get_resulting_id(self) -> str|None:
        return self.task.get_resulting_id()






import json
from typing import Iterable
from unittest import TestCase
from ....framework import (
    File, RunConfiguration, TestReport, SmallImageBuilder,
    DockerWebServiceController, IImageSource, RemotePublicImageSource,
    BrainBoxApi, BrainBoxTask
)
from ...common import check_if_its_sound, VOICEOVER_TEXT
from .settings import OpenTTSSettings
from pathlib import Path


class OpenTTSController(DockerWebServiceController[OpenTTSSettings]):
    def get_image_source(self) -> IImageSource:
        return RemotePublicImageSource(
            'synesthesiam/opentts:en',
            self.get_snake_case_name()
        )


    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 5500},
            mount_top_resource_folder=False,
        )

    def get_default_settings(self):
        return OpenTTSSettings()

    def create_api(self):
        from .api import OpenTTS
        return OpenTTS()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import OpenTTS

        result = api.execute(BrainBoxTask.call(OpenTTS).voiceover(VOICEOVER_TEXT))
        yield TestReport.last_call(api).href('voiceover').with_comment("Voiceover").result_is_file(File.Kind.Audio)
        result = api.open_file(result)
        check_if_its_sound(result.content, tc)




from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, INotebookableController, BrainBoxExtendedTask, MediaLibrary
)
from .settings import ResemblyzerSettings
from pathlib import Path


class ResemblyzerController(DockerWebServiceController[ResemblyzerSettings], INotebookableController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port: 8084}
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return ResemblyzerSettings()

    def create_api(self):
        from .api import Resemblyzer
        return Resemblyzer()


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Resemblyzer

        model_name = 'test'
        media_library_path = Path(__file__).parent / 'test_media_library.zip'
        prerequisites = Resemblyzer.upload_dataset(model_name, media_library_path, False)


        api.execute(BrainBoxExtendedTask(
            BrainBoxTask.call(Resemblyzer).train(model_name),
            prerequisite=prerequisites
        ))

        yield (TestReport
               .last_call(api)
               .href('train')
               .with_comment("Training the model with pre-uploaded resources")
               )

        media_library = MediaLibrary.read(media_library_path)

        first_time = True
        first_file = None
        for record in media_library.records:
            if record.tags['split'] == 'test':
                file = record.get_file()
                result = api.execute(BrainBoxTask.call(Resemblyzer).classify(file, model_name))
                if first_time:
                    first_file = file
                    yield TestReport.last_call(api).href('inference').with_comment("Running inference with trained model")
                first_time = False
                tc.assertEqual(record.tags['speaker'], result)

        api.execute(BrainBoxTask.call(Resemblyzer).distances(first_file,model_name))
        yield TestReport.last_call(api).href('distances')




DOCKERFILE = f'''
FROM python:3.8.20

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . /home/app/kaia

ENTRYPOINT ["python3","/home/app/kaia/main.py"]
'''

DEPENDENCIES = '''
anyio==4.4.0
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==2.4.1
async-lru==2.0.4
attrs==23.2.0
audioread==3.0.1
Babel==2.15.0
backcall==0.2.0
beautifulsoup4==4.12.3
bleach==6.1.0
blinker==1.8.2
certifi==2024.7.4
cffi==1.16.0
charset-normalizer==3.3.2
click==8.1.7
comm==0.2.2
contourpy==1.1.1
cycler==0.12.1
debugpy==1.8.2
decorator==5.1.1
defusedxml==0.7.1
exceptiongroup==1.2.1
executing==2.0.1
fastjsonschema==2.20.0
filelock==3.15.4
Flask==3.0.3
fonttools==4.53.1
fqdn==1.5.1
fsspec==2024.6.1
h11==0.14.0
httpcore==1.0.5
httpx==0.27.0
idna==3.7
importlib_metadata==8.0.0
importlib_resources==6.4.0
ipykernel==6.29.5
ipython==8.12.3
isoduration==20.11.0
itsdangerous==2.2.0
jedi==0.19.1
Jinja2==3.1.4
joblib==1.4.2
json5==0.9.25
jsonpointer==3.0.0
jsonschema==4.22.0
jsonschema-specifications==2023.12.1
jupyter-events==0.10.0
jupyter-lsp==2.2.5
jupyter_client==8.6.2
jupyter_core==5.7.2
jupyter_server==2.14.1
jupyter_server_terminals==0.5.3
jupyterlab==4.2.3
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.2
kiwisolver==1.4.5
lazy_loader==0.4
librosa==0.10.2.post1
llvmlite==0.41.1
MarkupSafe==2.1.5
matplotlib==3.7.5
matplotlib-inline==0.1.7
mistune==3.0.2
mpmath==1.3.0
msgpack==1.0.8
nbclient==0.10.0
nbconvert==7.16.4
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.1
notebook==7.2.1
notebook_shim==0.2.4
numba==0.58.1
numpy==1.24.4
nvidia-cublas-cu12==12.1.3.1
nvidia-cuda-cupti-cu12==12.1.105
nvidia-cuda-nvrtc-cu12==12.1.105
nvidia-cuda-runtime-cu12==12.1.105
nvidia-cudnn-cu12==8.9.2.26
nvidia-cufft-cu12==11.0.2.54
nvidia-curand-cu12==10.3.2.106
nvidia-cusolver-cu12==11.4.5.107
nvidia-cusparse-cu12==12.1.0.106
nvidia-nccl-cu12==2.20.5
nvidia-nvjitlink-cu12==12.5.82
nvidia-nvtx-cu12==12.1.105
overrides==7.7.0
packaging==24.1
pandas==2.0.3
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
pickleshare==0.7.5
pillow==10.4.0
pkgutil_resolve_name==1.3.10
platformdirs==4.2.2
pooch==1.8.2
prometheus_client==0.20.0
prompt_toolkit==3.0.47
psutil==6.0.0
ptyprocess==0.7.0
pure-eval==0.2.2
pycparser==2.22
Pygments==2.18.0
pyparsing==3.1.2
python-dateutil==2.9.0.post0
python-json-logger==2.0.7
pytz==2024.1
PyYAML==6.0.1
pyzmq==26.0.3
referencing==0.35.1
requests==2.32.3
Resemblyzer==0.1.4
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rpds-py==0.19.0
scikit-learn==1.3.2
scipy==1.10.1
Send2Trash==1.8.3
six==1.16.0
sniffio==1.3.1
soundfile==0.12.1
soupsieve==2.5
soxr==0.3.7
stack-data==0.6.3
sympy==1.12.1
terminado==0.18.1
threadpoolctl==3.5.0
tinycss2==1.3.0
tomli==2.0.1
torch==2.3.1
tornado==6.4.1
tqdm==4.66.4
traitlets==5.14.3
triton==2.3.1
types-python-dateutil==2.9.0.20240316
typing==3.7.4.3
typing_extensions==4.12.2
tzdata==2024.1
uri-template==1.3.0
urllib3==2.2.2
wcwidth==0.2.13
webcolors==24.6.0
webencodings==0.5.1
webrtcvad==2.0.10
websocket-client==1.8.0
Werkzeug==3.0.3
zipp==3.19.2
'''




from brainbox.framework import BrainBoxServer, ControllerRegistry, BrainBoxServiceSettings, Locator
import argparse
from pathlib import Path

parser = argparse.ArgumentParser()
parser.add_argument('-f', '--data-folder', required=True)
parser.add_argument('-p', '--port', default='8090')

if __name__ == '__main__':
    args = parser.parse_args()
    locator = Locator(Path(args.data_folder))
    registry = ControllerRegistry.discover()
    registry.locator = locator
    settings = BrainBoxServiceSettings(
        registry=registry,
        locator=locator,
        port=int(args.port)
    )
    server = BrainBoxServer(settings)
    server()



from .p01_installation_and_self_test import *
from .p02_simple_runs import *
from .p03_downloads import *
from .p04_uploads import *
from .p05_resources import *
from .p06_dependent_tasks import *
from .p07_without_api import *

def run_all(api: BrainBox.Api, test_case: TestCase|None = None):
    if test_case is None:
        test_case = TestCase()

    print('Uninstalling HelloBrainBox')
    api.controller_api.uninstall(HelloBrainBox, True)

    print('Installing HelloBrainBox')
    install(test_case, api)

    print("Self-testing HelloBrainBox")
    self_test(test_case, api)

    print("Running decider without parameter")
    run_without_parameters(test_case, api)

    print("Running decider with parameter")
    run_with_parameter(test_case, api)

    print("Test downloading files")
    downloads(test_case, api)

    print("Test uploading files")
    uploads(test_case, api)

    print("Resources")
    resources(test_case, api)

    print("Model installation")
    model_installation(test_case, api)

    print("Dependent tasks")
    dependent_tasks(test_case, api)

    print("Collector")
    collector(test_case, api)

    print("Running without API")
    pure_http(test_case, api)

    print("ALL GOOD")


import requests
from pathlib import Path
from ....framework import DockerWebServiceApi, FileLike, BrainBoxApi, MediaLibrary, ResourcePrerequisite, CombinedPrerequisite
from .settings import ResemblyzerSettings
from .controller import ResemblyzerController


class Resemblyzer(DockerWebServiceApi[ResemblyzerSettings, ResemblyzerController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)

    def __call__(self, file: FileLike.Type, model: str):
        return self.classify(file, model)


    def classify(self, file: str|Path|bytes, model: str):
        with FileLike(file, self.cache_folder) as file:
            reply = requests.post(
                f'http://{self.address}/classify/{model}',
                files=(
                    ('file', file),
                )
            )
            if reply.status_code != 200:
                raise ValueError(f"Resemblyzer threw an error\n{reply.text}")
            return reply.json()['speaker']


    def distances(self, file: FileLike.Type, model: str):
        with FileLike(file, self.cache_folder) as file:
            reply = requests.post(
                f'http://{self.address}/distances/{model}',
                files = (
                    ('file', file),
                )
            )
            if reply.status_code != 200:
                raise ValueError(f"Resemblyzer threw an error\n{reply.text}")
            return reply.json()


    def train(self, model: str):
        reply = requests.post(f'http://{self.address}/train/{model}')
        if reply.status_code != 200:
            raise ValueError(f"Resemblyzer threw an error\n{reply.text}")
        return reply.json()

    @staticmethod
    def upload_dataset_file(model: str, split: str, speaker: str, file: FileLike.Type) -> ResourcePrerequisite:
        fname = FileLike.get_name(file, True)
        return ResourcePrerequisite(
            Resemblyzer,
            f'datasets/{model}/{split}/{speaker}/{fname}',
            file
        )


    @staticmethod
    def delete_dataset(model: str) -> ResourcePrerequisite:
        return ResourcePrerequisite(
            Resemblyzer,
            f'datasets/{model}',
            None
        )

    @staticmethod
    def upload_dataset(model: str, media_library_path: Path, append_dataset: bool = False) -> CombinedPrerequisite:
        command = []
        media_library = MediaLibrary.read(media_library_path)
        if not append_dataset:
            command.append(Resemblyzer.delete_dataset(model))
        for record in media_library:
            command.append(Resemblyzer.upload_dataset_file(
                model,
                record.tags['split'],
                record.tags['speaker'],
                record.get_file()
            ))
        return CombinedPrerequisite(command)


    Controller = ResemblyzerController
    Settings = ResemblyzerSettings



from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, File, IModelDownloadingController, DownloadableModel
)
from .settings import YoloSettings
from .model import YoloModel
from pathlib import Path


class YoloController(DockerWebServiceController[YoloSettings], IModelDownloadingController):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return YoloModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            parameter,
            publish_ports={self.connection_settings.port:8084},
        )

    def get_default_settings(self):
        return YoloSettings()

    def create_api(self):
        from .api import Yolo
        return Yolo()

    def post_install(self):
        self.download_models(self.settings.models_to_downloads)


    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import Yolo

        api.execute(BrainBoxTask.call(Yolo).analyze(
            File.read(Path(__file__).parent / "test_anime_img.png"),
            "Fuyucchi/yolov8_animeface:yolov8x6_animeface.pt"
        ))

        yield TestReport.last_call(api).href('run')





DOCKERFILE = f"""
FROM python:3.11.11

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

WORKDIR /home/app

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

COPY . /home/app/

ENTRYPOINT ["python3", "/home/app/main.py"]
"""

DEPENDENCIES = """
    annotated-types==0.7.0
    anyio==4.5.2
    certifi==2019.11.28
    chardet==3.0.4
    charset-normalizer==3.4.0
    click==8.1.7
    contourpy==1.1.1
    cycler==0.12.1
    dbus-next
    exceptiongroup==1.2.2
    fastapi==0.115.3
    filelock==3.16.1
    fonttools==4.55.3
    fsspec==2024.10.0
    h11==0.14.0
    huggingface-hub==0.25.1
    idna==2.8
    importlib-resources==6.4.5
    jinja2==3.1.4
    kiwisolver==1.4.7
    loguru==0.7.2
    MarkupSafe==2.1.5
    matplotlib==3.7.5
    mpmath==1.3.0
    networkx==3.1
    numpy==1.24.4
    nvidia-cublas-cu12==12.1.3.1
    nvidia-cuda-cupti-cu12==12.1.105
    nvidia-cuda-nvrtc-cu12==12.1.105
    nvidia-cuda-runtime-cu12==12.1.105
    nvidia-cudnn-cu12==9.1.0.70
    nvidia-cufft-cu12==11.0.2.54
    nvidia-curand-cu12==10.3.2.106
    nvidia-cusolver-cu12==11.4.5.107
    nvidia-cusparse-cu12==12.1.0.106
    nvidia-nccl-cu12==2.20.5
    nvidia-nvjitlink-cu12==12.6.85
    nvidia-nvtx-cu12==12.1.105
    opencv-contrib-python-headless==4.10.0.82
    opencv-python==4.10.0.84
    packaging==24.2
    pandas==2.0.3
    pillow==10.4.0
    psutil==6.1.0
    py-cpuinfo==9.0.0
    pydantic==2.9.2
    pydantic-core==2.23.4
    pyparsing==3.1.4
    python-dateutil==2.9.0.post0
    python-multipart==0.0.17
    pytz==2024.2
    PyYAML==6.0.2
    requests==2.32.3
    requests-unixsocket==0.2.0
    scipy==1.10.1
    seaborn==0.13.2
    six==1.14.0
    sniffio==1.3.1
    starlette==0.41.3
    sympy==1.13.3
    torch==2.4.1
    torchvision==0.19.1
    tqdm==4.67.1
    triton==3.0.0
    typing-extensions==4.12.2
    tzdata==2024.2
    ultralytics==8.3.15
    ultralytics-thop==2.0.13
    urllib3==1.25.8
    uvicorn==0.32.0
    zipp==3.20.2
"""


import base64
import io
import os
import shutil
from pathlib import Path

from huggingface_hub import hf_hub_download
from loguru import logger

from fastapi import APIRouter, FastAPI, Request
from fastapi.responses import JSONResponse, PlainTextResponse
from ultralytics import YOLO

from validation_objects import LoadModelRequest
from validation_objects import CoordsFaceSquare
from PIL import Image
import traceback


class RecognizerApp:
    router = APIRouter()

    def __init__(self):
        self._similarity_threshold = 0.9
        self.model = None
        self.model_name = None
        self.data_directory = '/resources'
        self.models_directory = "/resources/models"
        self.hf_directory = '/resources/hf'
        os.makedirs(self.models_directory, exist_ok=True)
        os.makedirs(self.hf_directory, exist_ok=True)
        self.name_image_png = "image.png"

    def create_app(self):
        app = FastAPI()
        RecognizerApp.router.add_api_route("/", self.index, methods=["GET"])
        RecognizerApp.router.add_api_route("/post_image", self.post_image, methods=["POST"])
        RecognizerApp.router.add_api_route("/get_coordinates_faces", self.get_coordinates_faces, methods=["GET"])
        RecognizerApp.router.add_api_route("/load_model", self.load_model, methods=["POST"])
        RecognizerApp.router.add_api_route("/get_loaded_model", self.get_loaded_model, methods=["GET"])
        app.include_router(RecognizerApp.router)
        app.add_exception_handler(Exception, self.exception_handler)

        return app

    def index(self):
        return f"{type(self).__name__} is running"

    async def exception_handler(self, request: Request, exc: Exception):
        tb = traceback.format_exc()
        return PlainTextResponse(
            content=f"An internal error occurred:\n\n{tb}",
            status_code=500,
        )

    def _model_id_to_path(self, model_id: str):
        repo_id, model_filename = model_id.split(':')
        return os.path.join(self.models_directory, repo_id, model_filename)

    def _download_model(self, model_id: str):
        repo_id, model_filename = model_id.split(':')
        path_to_cache_model = hf_hub_download(repo_id=repo_id, filename=model_filename, cache_dir=self.hf_directory)
        path = self._model_id_to_path(model_id)
        os.makedirs(Path(path).parent, exist_ok=True)
        shutil.copy(path_to_cache_model, self._model_id_to_path(model_id))


    def _load_model_internal(self, model_id: str):
        path = self._model_id_to_path(model_id)
        if not os.path.isfile(path):
            self._download_model(model_id)
        self.model = YOLO(path)
        self.model_name = model_id
        try:
            self.model.to("cuda")
        except:
            self.model.to("cpu")
        return "OK"

    async def load_model(self, request: LoadModelRequest):
        return self._load_model_internal(request.model_id)

    def get_loaded_model(self):
        return JSONResponse(content={"name": self.model_name})

    async def post_image(self, request: Request):
        await self._save_img_base64_to_png(request)
        return "OK"

    def get_coordinates_faces(self) -> dict[str,list]:
        source = f"{self.data_directory}/{self.name_image_png}"
        results = self.model(source, stream=False)
        faces_coords = self._format_result(results)

        return {"objects": faces_coords}

    def _format_result(self, results) -> list:
        faces_coords = []
        for result in results:
            for box in result.boxes.xywh:
                x_center, y_center, w, h = round(box[0].item()), round(box[1].item()), round(box[2].item()), round(box[3].item())
                faces_coords.append({
                    "x_center": x_center,
                    "y_center": y_center,
                    "w": w,
                    "h": h,
                })
        return faces_coords

    async def _save_img_base64_to_png(self, request: Request):
        form = await request.form()
        img_base64 = form["image_base64"]
        bytes_img = base64.b64decode(img_base64.replace("data:image/jpeg;base64,", "").strip())
        bytes_io = io.BytesIO(bytes_img)
        bytes_io.seek(0)
        img = Image.open(bytes_io)
        img.save(f"{self.data_directory}/{self.name_image_png}", "PNG")


from typing import cast
from pathlib import Path

from ....framework import DownloadableModel, File, IController
from dataclasses import dataclass

@dataclass
class WD14TaggerModel(DownloadableModel):
    name: str
    file_name: str

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder('models')/'hub'/self.file_name

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        from .controller import WD14TaggerController
        from .api import WD14Tagger
        controller = cast(WD14TaggerController, controller)
        instance_id = controller.run(None)
        api: WD14Tagger = controller.find_api(instance_id)
        file = File.read(Path(__file__).parent/'image.png')
        for model in models:
            model = cast(WD14TaggerModel, model)
            api.interrogate(file, model=model.name)
        controller.stop(instance_id)





import subprocess


def run_espeak(data):
    text = data['text']
    text = '\n'.join(text)

    language = data.get('language', 'en-us')
    stress = data.get('stress', False)

    command = [
        '/home/app/.local/bin/espeak-phonemizer',
        '-v',
        language,
        '-p',
        '#',
        '-w',
        '$'
    ]

    if not stress:
        command.append('--no-stress')

    result = subprocess.check_output(
        command,
        input=text,
        text=True)

    result = [[s.split('#') for s in line.split('$')] for line in result.split('\n')[:-1]]
    return result


import os
from pathlib import Path
import shutil
from .file_system import IFileSystem


class LocalFileSystem(IFileSystem):
    def upload_file(self, local_path: Path, remote_path: str):
        os.makedirs(Path(remote_path).parent, exist_ok=True)
        shutil.copy(local_path, remote_path)

    def download_file(self, remote_path: str, local_path: Path):
        shutil.copy(remote_path, local_path)

    def create_empty_folder(self, path: str|Path):
        os.makedirs(path, exist_ok=True)

    def is_file(self, remote_path: str):
        return Path(remote_path).is_file()

    def is_dir(self, remote_path: str):
        return Path(remote_path).is_dir()

    def delete_file_or_folder(self, remote_path: str):
        if self.is_file(remote_path):
            os.unlink(remote_path)
        elif self.is_dir(remote_path):
            shutil.rmtree(remote_path)



from ....job_processing import OperatorLogItem

def create_operator_log_page(log: list[OperatorLogItem]):
    html = []
    for item in log:
        html.append('<tr>')
        html.append(f'<td>{item.timestamp}</td>')
        html.append(f'<td>{item.level.name}</td>')
        html.append(f'<td>{item.id}</td>')
        html.append(f'<td>{item.event}</td>')
        if item.error is not None:
            error = item.error.replace('\n','<br>')
            html.append(f'<td><tt>{error}</tt></td>')
        html.append('</tr>')
    return '<html><body><table border=1 cellspacing=1>'+'\n'.join(html)+'</table></body></html>'




from typing import Iterable, Optional
import os
import requests
from ....framework import File, DockerWebServiceApi, FileIO, Loc, LocalExecutor, CombinedPrerequisite, IPrerequisite
from pathlib import Path
from .controller import TortoiseTTSController
from .settings import TortoiseTTSSettings
from .upload_prerequisite import TortoiseTTSUploadPrerequisite





class TortoiseTTS(DockerWebServiceApi[TortoiseTTSSettings, TortoiseTTSController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def dub(self, text: str, voice: str, count = 3):
        result = requests.post(f'http://{self.address}/dub', json=dict(output_file_name = self.current_job_id, text=text, voice=voice, count=count))
        if result.status_code == 500:
            raise ValueError(f"TortoiseTTS server returned {result.status_code}\n{result.text}")
        files = []
        for i, result in enumerate(result.json()):
            path = self.controller.resource_folder('stash')/result
            files.append(File(
                self.current_job_id+f'.output.{i}.wav',
                FileIO.read_bytes(path),
                File.Kind.Audio
            ))
            os.unlink(path)
        return files


    @staticmethod
    def export_voice(
            voice: str,
            files: Optional[Iterable[Path]] = None,
            folder: Optional[Path] = None
    ) -> CombinedPrerequisite:
        if voice is None:
            raise ValueError("Voice cannot be None")

        if folder is not None:
            files = [folder/c for c in os.listdir(folder)]
        elif files is not None:
            files = list(files)
        else:
            raise ValueError(f"`files` and `folder` were both null")

        command = [TortoiseTTSUploadPrerequisite(voice, file) for file in files]
        return CombinedPrerequisite(command)


    Controller = TortoiseTTSController
    Settings = TortoiseTTSSettings


from brainbox import BrainBox
from brainbox.deciders import HelloBrainBox
from unittest import TestCase


def install(test_case: TestCase, api: BrainBox.Api):
    """
    # Python's API

    This is the simplest way to access the BrainBox functionality.
    To use it, simply install BrainBox to your project with `pip install brainbox`

    To execute the following code, you need to run BrainBox and keep it running.

    Note, that the code below is also part of the tests and is located at `/doc`.
    In this documentation, regions from this test are copy-pasted,
    along with asserts for better understanding of returned values,
    so `test_case` represents a `unittest` TestCase instance.

    ## Basic techniques

    These are the operations you use for all the deciders all the time you work with BrainBox.

    ### Create an API object

    ```python
    from brainbox import BrainBox

    api = BrainBox.Api("127.0.0.1:8090")
    api.wait(1)
    ```

    This will create an API object. `wait` will try to connect to the endpoint
    for 1 second, and raises if unsuccessful.

    ### Install the decider

    We have a `HelloBrainBox` decider, that doesn't do anything useful,
    but accumulates all the features BrainBox has for deciders.
    Let's install the decider:
    """

    from brainbox.deciders import HelloBrainBox

    report = api.controller_api.install(HelloBrainBox)
    test_case.assertIsNotNone(report)

    """
    Note that the container is not pulled, but build on your machine, 
    and this is also the case for other deciders. 
    This is resource-intensive, in terms of bandwidth and time,
    but allows you to modify the containers quickly 
    if you require some model's functionality that wasn't implemented.
    Also it allows BrainBox to work completely independently,
    with the source code only, without reliance to the external 
    docker repositories; and finally, it's much easier to check what 
    these containers are doing.
        
    Returned `report` can largely be ignored, it's only needed for
    GUI purposes.
    """

def self_test(test_case: TestCase, api: BrainBox.Api):
    """
    ### Run a self-test

    The self-tests for the deciders run the most important endpoints,
    and ensure their correct work.
    Also, they provide a report where you can see the inputs and outputs
    for the endpoint, so it is documentation of some sort.
    This report can be viewed as an HTML page.

    This will run a self-test and opens the result in the web-browser.
    """

    import requests, webbrowser, tempfile
    from pathlib import Path

    api.controller_api.self_test(HelloBrainBox)
    self_test_report = requests.get(f'http://{api.address}/html/controllers/self_test_report/HelloBrainBox').text
    test_case.assertIsInstance(self_test_report, str)

    test_case_test_path = Path(tempfile.gettempdir()) / 'test_report.html'
    with open(test_case_test_path, 'w', encoding='utf-8') as file:
        file.write(self_test_report)

    """
    
    and then:
    
    ```python
    webbrowser.open('file://'+str(test_case_test_path))
    ```
    
    While self-test reports provide some understanding of the endpoints,
    I recommend reading the self-tests code: all the endpoints are called 
    in the same way you will call them from your Python code.
    Self-tests, building containers and other container-related things
    are performed by `Controller` classes. They are usually located
    within deciders classes, e.g. `HelloBrainBox.Controller`
    """


from typing import *
from dataclasses import dataclass
from pathlib import Path
from ....framework import DownloadableModel, IController
from ...common import download_file


@dataclass
class PiperTrainingModel(DownloadableModel):
    name: str
    url: str

    def get_local_location(self, controller: IController) -> Path:
        return controller.resource_folder('base_models')/self.name

    @classmethod
    def download(cls, models: list['DownloadableModel'], controller: IController):
        for model in models:
            model = cast(PiperTrainingModel, model)
            download_file(model.url, controller.resource_folder('base_models')/model.name)



import cv2
import json
import os
import shutil
from PIL import Image
from settings import AnalysisSettings
from frame import Frame
from layer_capture import layer_capture
from layer_bufferer import layer_bufferer
from layer_semantic_comparator import layer_comparator

class VideoProcessorApp:
    def __init__(self):
        self.input_directory = '/resources/input'
        self.output_directory = '/resources/output/images'
        self.timestamps = '/resources/output/frames.jsonl'
        self.settings_file = '/resources/settings.json'
        os.makedirs(self.output_directory, exist_ok=True)


    def run(self):
        with open(self.settings_file,'r') as file:
            settings = AnalysisSettings(**json.load(file))

        source = layer_capture(settings)
        if settings.buffer_by_laplacian_in_ms is not None:
            source = layer_bufferer(source, settings.buffer_by_laplacian_in_ms, 'laplacian')
        if settings.add_semantic_comparator:
            source = layer_comparator(source)

        with open(self.timestamps, 'w') as stream:
            for index, frame in enumerate(source):
                if settings.max_produced_frames is not None and index>=settings.max_produced_frames:
                    print(f"Interrupting at {index} >= {settings.max_produced_frames}")
                    break

                frame.pil_image.save(os.path.join(self.output_directory, frame.filename), format="PNG")
                data = frame.__dict__
                del data['frame']
                del data['pil_image']
                stream.write(json.dumps(data))
                stream.write('\n')
                stream.flush()






import os
from pathlib import Path
from rhasspyserver_hermes.train import sentences_to_graph
from rhasspyasr_kaldi.train import train, LanguageModelType
import rhasspynlu
from collections import defaultdict
from utils import kaldi_path, profile_path

class Trainer:
    def __init__(self,
                 path: Path,
                 sentences: str,
                 language: str,
                 custom_words: str = '',
                 ):
        self.path = path
        self.sentences = sentences
        self.custom_words = custom_words
        self.language = language


    def _build_and_save_graph(self):
        graph, slot_replacements = sentences_to_graph(
            sentences_dict={str(self.path / 'sentences.ini'): self.sentences},
            slots_dirs=[
                self.path / 'slots',
                profile_path / self.language/'slots',
            ],
            slot_programs_dirs=[
                self.path / 'slot_programs',
                profile_path / self.language/'slot_programs',
            ],
            language='en',
            word_transform=str.lower
        )

        with open(self.path / 'intent_graph.pickle.gz', 'wb') as file:
            rhasspynlu.graph_to_gzip_pickle(graph, file)

        return graph

    def _get_pronunciations(self):
        custom_words_path = self.path/'kaldi_custom_words.txt'
        with open(custom_words_path, 'w') as file:
            file.write(self.custom_words)

        dictionaries = [
            self.path / 'kaldi/base_dictionary.txt',
            custom_words_path,
        ]
        pronunciations = defaultdict(list)

        for dict_path in dictionaries:
            with open(dict_path, "r") as base_dict_file:
                dict_content = {}
                rhasspynlu.g2p.read_pronunciations(base_dict_file, word_dict=dict_content)
                for word in dict_content:
                    pronunciations[word].extend(dict_content[word])

        return pronunciations

    def _get_frequent_words(self):
        with open(profile_path / self.language/'frequent_words.txt', 'r') as file:
            frequent_words = [s.strip() for s in file][:100]
        return frequent_words


    def train(self):
        for key, value in ENVS.items():
            os.environ[key] = value

        train(
            graph=self._build_and_save_graph(),
            pronunciations=self._get_pronunciations(),
            model_dir=self.path / 'kaldi/model',
            graph_dir=self.path / 'kaldi/model/graph',
            dictionary=self.path / 'kaldi/dictionary.txt',
            language_model=self.path / 'kaldi/language_model.txt',
            dictionary_word_transform=str.lower,
            g2p_model=self.path / 'kaldi/g2p.fst',
            g2p_word_transform=None,
            missing_words_path=self.path / 'kaldi_unknown_words.txt',
            vocab_path=None,
            language_model_fst=None,
            base_language_model_fst=self.path / 'kaldi/base_language_model.fst',
            base_language_model_weight=0,
            mixed_language_model_fst=self.path / 'kaldi/mixed.fst',
            balance_counts=True,
            kaldi_dir=kaldi_path,
            language_model_type=LanguageModelType.TEXT_FST,
            spn_phone='SPN',
            sil_phone='SIL',
            eps='<eps>',
            allow_unknown_words=False,
            frequent_words=self._get_frequent_words(),
            unk='<unk>',
            sil='<sil>',
            unk_nonterm='#nonterm:unk',
            unk_prob=1e-05,
            sil_prob=0.5,
            unknown_token='<unk>',
            max_unk_words=8,
            cancel_word='',
            cancel_prob=0.01
        )

        unknown_words = {}
        with open(self.path/'kaldi_unknown_words.txt') as file:
            for line in file:
                try:
                    first_space = line.index(' ')
                except ValueError:
                    pass
                unknown_words[line[:first_space].strip()] = line[first_space:].strip()
        return dict(status='OK', unknown_words=unknown_words)



ENVS = {
    'SUPERVISOR_GROUP_NAME': 'speech_to_text',
    'HOSTNAME': '6c46571adf40',
    'RHASSPY_PROFILE_DIR': '/profiles/en',
    'PWD': '/',
    'HOME': '/',
    'LANG': 'C.UTF-8',
    'VIRTUAL_ENV': '/usr/lib/rhasspy/.venv',
    'KALDI_DIR': '/usr/lib/rhasspy/.venv/lib/kaldi',
    'PYTHONPATH': '/usr/lib/rhasspy/rhasspy-asr-kaldi-hermes:/usr/lib/rhasspy:/usr/lib/rhasspy/rhasspy-wake-snowboy-hermes:/usr/lib/rhasspy/rhasspy-wake-raven-hermes:/usr/lib/rhasspy/rhasspy-wake-raven:/usr/lib/rhasspy/rhasspy-wake-precise-hermes:/usr/lib/rhasspy/rhasspy-wake-porcupine-hermes:/usr/lib/rhasspy/rhasspy-wake-pocketsphinx-hermes:/usr/lib/rhasspy/rhasspy-tts-wavenet-hermes:/usr/lib/rhasspy/rhasspy-tts-larynx-hermes:/usr/lib/rhasspy/rhasspy-tts-cli-hermes:/usr/lib/rhasspy/rhasspy-supervisor:/usr/lib/rhasspy/rhasspy-speakers-cli-hermes:/usr/lib/rhasspy/rhasspy-snips-nlu-hermes:/usr/lib/rhasspy/rhasspy-snips-nlu:/usr/lib/rhasspy/rhasspy-silence:/usr/lib/rhasspy/rhasspy-server-hermes:/usr/lib/rhasspy/rhasspy-remote-http-hermes:/usr/lib/rhasspy/rhasspy-rasa-nlu-hermes:/usr/lib/rhasspy/rhasspy-profile:/usr/lib/rhasspy/rhasspy-nlu-hermes:/usr/lib/rhasspy/rhasspy-nlu:/usr/lib/rhasspy/rhasspy-microphone-pyaudio-hermes:/usr/lib/rhasspy/rhasspy-microphone-cli-hermes:/usr/lib/rhasspy/rhasspy-homeassistant-hermes:/usr/lib/rhasspy/rhasspy-hermes:/usr/lib/rhasspy/rhasspy-fuzzywuzzy-hermes:/usr/lib/rhasspy/rhasspy-fuzzywuzzy:/usr/lib/rhasspy/rhasspy-dialogue-hermes:/usr/lib/rhasspy/rhasspy-asr-vosk-hermes:/usr/lib/rhasspy/rhasspy-asr-pocketsphinx-hermes:/usr/lib/rhasspy/rhasspy-asr-pocketsphinx:/usr/lib/rhasspy/rhasspy-asr-kaldi-hermes:/usr/lib/rhasspy/rhasspy-asr-kaldi:/usr/lib/rhasspy/rhasspy-asr-deepspeech-hermes:/usr/lib/rhasspy/rhasspy-asr-deepspeech:/usr/lib/rhasspy/rhasspy-asr:',
    'TERM': 'xterm',
    'RHASSPY_PROFILE': 'en',
    'SHLVL': '2',
    'APP_DIR': '/usr/lib/rhasspy',
    'LD_LIBRARY_PATH': '/usr/lib/rhasspy/.venv/lib:',
    'SUPERVISOR_PROCESS_NAME': 'speech_to_text',
    'PATH': '/usr/lib/rhasspy/rhasspy-wake-snowboy-hermes/bin:/usr/lib/rhasspy/rhasspy-wake-raven-hermes/bin:/usr/lib/rhasspy/rhasspy-wake-precise-hermes/bin:/usr/lib/rhasspy/rhasspy-wake-porcupine-hermes/bin:/usr/lib/rhasspy/rhasspy-wake-pocketsphinx-hermes/bin:/usr/lib/rhasspy/rhasspy-tts-wavenet-hermes/bin:/usr/lib/rhasspy/rhasspy-tts-larynx-hermes/bin:/usr/lib/rhasspy/rhasspy-tts-cli-hermes/bin:/usr/lib/rhasspy/rhasspy-speakers-cli-hermes/bin:/usr/lib/rhasspy/rhasspy-snips-nlu-hermes/bin:/usr/lib/rhasspy/rhasspy-server-hermes/bin:/usr/lib/rhasspy/rhasspy-remote-http-hermes/bin:/usr/lib/rhasspy/rhasspy-rasa-nlu-hermes/bin:/usr/lib/rhasspy/rhasspy-nlu-hermes/bin:/usr/lib/rhasspy/rhasspy-microphone-pyaudio-hermes/bin:/usr/lib/rhasspy/rhasspy-microphone-cli-hermes/bin:/usr/lib/rhasspy/rhasspy-homeassistant-hermes/bin:/usr/lib/rhasspy/rhasspy-fuzzywuzzy-hermes/bin:/usr/lib/rhasspy/rhasspy-dialogue-hermes/bin:/usr/lib/rhasspy/rhasspy-asr-vosk-hermes/bin:/usr/lib/rhasspy/rhasspy-asr-pocketsphinx-hermes/bin:/usr/lib/rhasspy/rhasspy-asr-kaldi-hermes/bin:/usr/lib/rhasspy/rhasspy-asr-deepspeech-hermes/bin:/usr/lib/rhasspy/.venv/bin:/usr/lib/rhasspy/.venv/bin:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin',
    'RHASSPY_BASE_DIR': '/usr/lib/rhasspy',
    'DEBIAN_FRONTEND': 'noninteractive',
    'SUPERVISOR_ENABLED': '1',
    '_': '/usr/lib/rhasspy/.venv/bin/python3'
}



from .decider_instance_key import DeciderInstanceKey
from dataclasses import dataclass
from datetime import datetime

@dataclass
class JobForPlanner:
    id: str
    decider: str
    decider_parameters: str|None
    received_timestamp: datetime
    assigned: bool
    ordering_token: str|None

    def get_decider_instance_key(self):
        return DeciderInstanceKey(self.decider, self.decider_parameters)


from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from .settings import WD14TaggerSettings
from .model import WD14TaggerModel
from pathlib import Path
import os

class WD14TaggerController(
    DockerWebServiceController[WD14TaggerSettings],
    IModelDownloadingController,
    INotebookableController
):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_downloadable_model_type(self) -> type[DownloadableModel]:
        return WD14TaggerModel

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        config = RunConfiguration(
            publish_ports={self.settings.connection.port: 8084},
            mount_resource_folders={'models': '/home/app/.cache/huggingface'},
        )
        if self.settings.cpu_share is not None:
            cpu_count = os.cpu_count()
            config.custom_flags=[f'--cpus={self.settings.cpu_share*cpu_count}']
        return config

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()

    def get_default_settings(self):
        return WD14TaggerSettings()

    def create_api(self):
        from .api import WD14Tagger
        return WD14Tagger()

    def post_install(self):
        self.download_models(self.settings.models_to_download)

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import WD14Tagger
        file = File.read(Path(__file__).parent / 'image.png')

        model = self.settings.models_to_download[0].name
        api.execute(BrainBoxTask.call(WD14Tagger).interrogate(image=file, model=model))
        yield TestReport.last_call(api)

        api.execute(BrainBoxTask.call(WD14Tagger).tags(model=model, count=10))
        yield TestReport.last_call(api)



DOCKERFILE = f'''
FROM python:3.11.11

{SmallImageBuilder.APT_INSTALL('ffmpeg libsm6 libxext6')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

{SmallImageBuilder.GIT_CLONE_AND_RESET(
    "https://github.com/corkborg/wd14-tagger-standalone", 
    "/home/app/wd14-tagger-standalone",
    "f1114c877ed6d1b1311f7e485ec0466d5064eb0c",
    install=False
)}

COPY . /home/app/wd14-tagger-standalone

WORKDIR /home/app/wd14-tagger-standalone

RUN touch tagger/__init__.py

RUN pip install -e .

ENTRYPOINT ["python3","/home/app/wd14-tagger-standalone/main.py"]
'''

DEPENDENCIES = '''
anyio==4.6.2.post1
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==2.4.1
async-lru==2.0.4
attrs==24.2.0
babel==2.16.0
beautifulsoup4==4.12.3
bleach==6.1.0
blinker==1.8.2
certifi==2024.8.30
cffi==1.17.1
charset-normalizer==3.4.0
click==8.1.7
coloredlogs==15.0.1
comm==0.2.2
debugpy==1.8.7
decorator==5.1.1
deepdanbooru==1.0.2
defusedxml==0.7.1
executing==2.1.0
fastjsonschema==2.20.0
filelock==3.16.1
Flask==3.0.3
flatbuffers==24.3.25
fqdn==1.5.1
fsspec==2024.9.0
h11==0.14.0
httpcore==1.0.6
httpx==0.27.2
huggingface-hub==0.22.2
humanfriendly==10.0
idna==3.10
imageio==2.36.0
ipykernel==6.29.5
ipython==8.28.0
isoduration==20.11.0
itsdangerous==2.2.0
jedi==0.19.1
Jinja2==3.1.4
json5==0.9.25
jsonpointer==3.0.0
jsonschema==4.23.0
jsonschema-specifications==2024.10.1
jupyter-events==0.10.0
jupyter-lsp==2.2.5
jupyter_client==8.6.3
jupyter_core==5.7.2
jupyter_server==2.14.2
jupyter_server_terminals==0.5.3
jupyterlab==4.2.5
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.3
lazy_loader==0.4
MarkupSafe==3.0.2
matplotlib-inline==0.1.7
mistune==3.0.2
mpmath==1.3.0
nbclient==0.10.0
nbconvert==7.16.4
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.4.1
notebook==7.2.2
notebook_shim==0.2.4
numpy==1.26.4
onnxruntime==1.17.3
opencv-python==4.9.0.80
overrides==7.7.0
packaging==24.1
pandas==2.2.2
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
Pillow==10.0.1
platformdirs==4.3.6
prometheus_client==0.21.0
prompt_toolkit==3.0.48
protobuf==5.28.2
psutil==6.1.0
ptyprocess==0.7.0
pure_eval==0.2.3
pycparser==2.22
Pygments==2.18.0
python-dateutil==2.9.0.post0
python-json-logger==2.0.7
pytz==2024.2
PyYAML==6.0.2
pyzmq==26.2.0
referencing==0.35.1
requests==2.32.3
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rpds-py==0.20.0
scikit-image==0.24.0
scipy==1.14.1
Send2Trash==1.8.3
six==1.16.0
sniffio==1.3.1
soupsieve==2.6
stack-data==0.6.3
sympy==1.13.3
terminado==0.18.1
tifffile==2024.9.20
tinycss2==1.3.0
tornado==6.4.1
tqdm==4.66.5
traitlets==5.14.3
types-python-dateutil==2.9.0.20241003
typing_extensions==4.12.2
tzdata==2024.2
uri-template==1.3.0
urllib3==2.2.3
wcwidth==0.2.13
webcolors==24.8.0
webencodings==0.5.1
websocket-client==1.8.0
Werkzeug==3.0.4
'''


from ....framework import DockerWebServiceApi, File, FileLike, CacheUploadPrerequisite
import requests
from .settings import HelloBrainBoxSettings
from .controller import HelloBrainBoxController
from .model import HelloBrainBoxModel
import json


class HelloBrainBox(DockerWebServiceApi[HelloBrainBoxSettings, HelloBrainBoxController]):
    def __init__(self, address: str|None = None):
        super().__init__(address)


    def json(self, argument: str):
        result = requests.post(
            self.endpoint('/decide'),
            json = dict(
                argument=argument,
            )
        )
        if result.status_code!=200:
            raise ValueError(f"Endpoint /decide returned unexpected status code {result.status_code}\n{result.text}")
        return result.json()

    def file(self, argument: str):
        json_data = self.json(argument)
        return File(self.current_job_id+'.output.json', json.dumps(json_data), File.Kind.Json)

    def file_length(self, file_like: FileLike.Type):
        with FileLike(file_like, self.cache_folder) as stream:
            return len(stream.read())

    def resources(self):
        result = requests.get(self.endpoint('/resources'))
        if result.status_code!=200:
            raise ValueError(f"Endpoint /resources returned unexpected status code {result.status_code}\n{result.text}")
        return result.json()


    @staticmethod
    def file_upload(file: File):
        return CacheUploadPrerequisite(file.content, file.name)


    Settings = HelloBrainBoxSettings
    Controller = HelloBrainBoxController
    Model = HelloBrainBoxModel


import argparse
import json
import subprocess
import sys
from pathlib import Path
import whisperx
import torch

def append_dict_to_json_array(file_path, data):
    with open(file_path, 'w', encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=4)

def load_models(args, device):
    model = whisperx.load_model(args.model, device=device, compute_type=args.compute_type,
                                download_root="/resources/models/")
    model_a, metadata = whisperx.load_align_model(
        language_code='ru',
        device=device
    )
    diarize_model = whisperx.diarize.DiarizationPipeline(
        use_auth_token=args.hf_token,
        device=device
    )

    return model, diarize_model, model_a, metadata

def handle_video(args, input_dir, output_dir):
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model, diarize_model, model_a, metadata = load_models(args, device)
    for video_path in input_dir.rglob('*.mp4'):
        audio = whisperx.load_audio(str(video_path))
        result = model.transcribe(audio, language=args.language, batch_size=16)
        result = whisperx.align(
            result["segments"],
            model_a,
            metadata,
            audio,
            device,
            return_char_alignments=False
        )
        diarize_segments = diarize_model(audio)
        result = whisperx.assign_word_speakers(diarize_segments, result)
        for segment in result["segments"]:
            segment.pop('words')
        result_with_meta = {
            "file": video_path.name,
            "segments": result["segments"]
        }
        append_dict_to_json_array(output_dir / 'output.json', result_with_meta)

def create_parser():
    parser = argparse.ArgumentParser(
        description="Video file processing with transcription and details"
    )

    parser.add_argument('-n', '--notebook', action='store_true')
    parser.add_argument('-i', '--input_dir', type=str, default='/resources/files')
    parser.add_argument('-o', '--output_dir', type=str, default='/resources/output')
    parser.add_argument('-m', '--model', type=str, choices=['base', 'large-v2'], default='base')
    parser.add_argument('--download_root', type=str, default='/resources/models')
    parser.add_argument('-t', '--compute_type', type=str, default='float32')
    parser.add_argument('-l', '--language', type=str, default='en')

    parser.add_argument('--hf_token', type=str)

    return parser

if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()

    if args.notebook:
        subprocess.call([sys.executable, '-m', 'notebook', '--allow-root', '--port', '8899', '--ip', '0.0.0.0',
                         "--NotebookApp.token=''"], cwd='/repo')
        exit(0)

    input_dir, output_dir = Path(args.input_dir), Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)
    handle_video(args, input_dir, output_dir)


import os
import shutil
import traceback
from pathlib import Path
from .steps import IStepFactory, IStep
from yo_fluq import *
import dill

def write_dill(obj, path):
    with open(path, 'wb') as f:
        dill.dump(obj, f)

def read_dill(path):
    with open(path, 'rb') as f:
        return dill.load(f)


class Flow:
    def __init__(self,
                 folder: Path|str,
                 steps: list[IStep|IStepFactory]|None = None,
                 prehistory: list|None = None
                 ):
        self.folder = Path(folder)
        if prehistory is None:
            prehistory = []
        self.prehistory = prehistory
        if steps is not None:
            self.steps = []
            for i, s in enumerate(steps):
                if isinstance(s, IStep):
                    self.steps.append(s)
                elif isinstance(s, IStepFactory):
                    self.steps.append(s.create_step())
                else:
                    raise ValueError(f"Erroneous step #{i}: must be IStep or IStepFactory, but was {type(s)}")
        else:
            self.steps = None

    def reset(self):
        shutil.rmtree(self.folder, ignore_errors=True)

    def _run_stages(self, iteration_index, history):
        current = []
        for stage_index, stage in enumerate(self.steps):
            name = stage.get_name()
            if name is None:
                name = type(stage).__name__
            print(f"{len(current)} records -> Start {name}, {stage_index}/{len(self.steps)} -> ", end='')
            full = None
            short = None
            try:
                full = stage.process(self.prehistory+history, current)
                short = stage.shorten(full)
                report = dict(index=stage_index, name=name, input=current, full_output=full, short_output=short, error = None)
                write_dill(report, self.folder/f'fulls/{iteration_index}_{stage_index}')
                summary = stage.summarize(full)
                if summary is not None:
                    summary = f' ({summary})'
                else:
                    summary = ''
                current = short
                print(f'{len(current)} records{summary}')
            except:
                print()
                report = dict(index=stage_index, name=name, input=current, full_output=full, short_output=short, error=traceback.format_exc())
                write_dill(report, self.folder / f'fulls/{iteration_index}_{stage_index}')
                raise

        write_dill(current, self.folder / f'results/{iteration_index}')
        return current


    def run(self, iterations):
        os.makedirs(self.folder / 'fulls', exist_ok=True)
        os.makedirs(self.folder / 'results', exist_ok=True)
        existing_files = Query.folder(self.folder/'results').to_list()
        existing = Query.en(existing_files).select_many(read_dill).to_list()
        print(f'TOTAL {len(existing)}')
        if len(existing_files) == 0:
            existing_index = -1
        else:
            existing_index = Query.en(existing_files).select(lambda z: int(z.name)).max()
        for index in range(existing_index+1, existing_index+1+iterations):
            result = self._run_stages(index, existing)
            existing.extend(result)
            print(f'TOTAL {len(existing)}')




    def read_iterations_count(self):
        index = 0
        while True:
            path = self.folder / f'results/{index}'
            if not path.is_file():
                return index-1
            index+=1


    def read_flatten(self):
        return Query.folder(self.folder/'results').where(lambda z: z.is_file()).order_by(lambda z: int(z.name)).select_many(read_dill).to_list()

    def read_debug_report(self, iteration: int, stage: int):
        return read_dill(self.folder/f'fulls/{iteration}_{stage}')





import flask
import shutil
import os
import pickle
import uuid
import traceback

from classifier import *

DATASETS_PATH = Path('/resources/datasets')
MODELS_FOLDER = Path('/resources/models')

class ResemblyzerServer:
    def __init__(self):
        self.wav_processor = WavProcessor()
        self.models_cache: dict[str, Model] = {}


    def __call__(self):
        app = flask.Flask(__name__)
        app.add_url_rule('/train/<model>', view_func=self.train, methods=['POST'])
        app.add_url_rule('/classify/<model>', view_func=self.classify, methods=['POST'])
        app.add_url_rule('/distances/<model>', view_func=self.distances, methods=['POST'])
        app.add_url_rule('/', view_func=self.index, methods=['GET'])
        app.run('0.0.0.0',8084)

    def _get_model(self, model):
        if model not in self.models_cache:
            with open(MODELS_FOLDER / model, 'rb') as file:
                self.models_cache[model] = Model(**pickle.load(file))
        return self.models_cache[model]

    def _get_embedding(self):
        path = Path('/resources') / str(uuid.uuid4())
        file = flask.request.files['file']
        file.save(path)
        embedding = self.wav_processor.get_encoding(path)
        os.unlink(path)
        return embedding

    def classify(self, model: str):
        try:
            embedding = self._get_embedding()
            winner = self._get_model(model).compute_winner(embedding)
            if winner is None:
                raise ValueError("Undef winner")
            return flask.jsonify(dict(speaker=winner))
        except:
            return traceback.format_exc(), 500

    def distances(self, model: str):
        try:
            embedding = self._get_embedding()
            result = self._get_model(model).compute_full(embedding)
            return flask.jsonify(result)
        except:
            return traceback.format_exc(), 500

    def index(self):
        return 'OK'


    def train(self, model: str):
        try:
            train = self.wav_processor.get_encodings(DATASETS_PATH/model/'train')
            model_instance = Model(train)

            test = self.wav_processor.get_encodings(DATASETS_PATH/model/'test')
            model_instance.evaluate(test)

            os.makedirs(MODELS_FOLDER, exist_ok=True)
            with open(MODELS_FOLDER/model,'wb') as file:
                pickle.dump(model_instance.__dict__, file)
            stats_df = model_instance.stats
            if stats_df is not None:
                stats_df = {column: list(model_instance.stats[column]) for column in stats_df}
            return flask.jsonify(dict(accuracy=model_instance.accuracy, stats = stats_df))
        except:
            return traceback.format_exc(), 500















from typing import Iterable
from unittest import TestCase
from ....framework import (
    RunConfiguration, TestReport, SmallImageBuilder, IImageBuilder, DockerWebServiceController,
    BrainBoxApi, BrainBoxTask, FileIO, INotebookableController, IModelDownloadingController, DownloadableModel,
    File
)
from .settings import ResembleEnhanceSettings
from ...common import check_if_its_sound
from pathlib import Path


class ResembleEnhanceController(
    DockerWebServiceController[ResembleEnhanceSettings],
    INotebookableController,
):
    def get_image_builder(self) -> IImageBuilder|None:
        return SmallImageBuilder(
            Path(__file__).parent/'container',
            DOCKERFILE,
            DEPENDENCIES.split('\n'),
        )

    def get_service_run_configuration(self, parameter: str|None) -> RunConfiguration:
        if parameter is not None:
            raise ValueError(f"`parameter` must be None for {self.get_name()}")
        return RunConfiguration(
            publish_ports={self.connection_settings.port:8080},
            mount_resource_folders={
                'uploads' : '/uploads',
                'outputs' : '/outputs',
                'models' : '/home/app/.local/lib/python3.11/site-packages/resemble_enhance/model_repo'
            },
        )

    def get_notebook_configuration(self) -> RunConfiguration|None:
        return self.get_service_run_configuration(None).as_notebook_service()



    def get_default_settings(self):
        return ResembleEnhanceSettings()

    def create_api(self):
        from .api import ResembleEnhance
        return ResembleEnhance()

    def _self_test_internal(self, api: BrainBoxApi, tc: TestCase) -> Iterable:
        from .api import ResembleEnhance

        input = File.read(Path(__file__).parent/'sample_10db.wav')
        result = api.execute(BrainBoxTask.call(ResembleEnhance).process(input))
        for file in result:
            check_if_its_sound(api.open_file(file).content, tc)

        yield (
            TestReport
            .last_call(api)
            .href('10db')
            .result_is_array_of_files(File.Kind.Audio)
        )


DOCKERFILE = f"""
FROM python:3.11.11

{SmallImageBuilder.APT_INSTALL('git-lfs')}

{{{SmallImageBuilder.ADD_USER_PLACEHOLDER}}}

{{{SmallImageBuilder.PIP_INSTALL_PLACEHOLDER}}}

WORKDIR /home/app

COPY . /home/app



ENTRYPOINT ["python3", "/home/app/main.py"]
"""


DEPENDENCIES = """
aiofiles==23.2.1
altair==5.5.0
annotated-types==0.7.0
antlr4-python3-runtime==4.9.3
anyio==4.8.0
argon2-cffi==23.1.0
argon2-cffi-bindings==21.2.0
arrow==1.3.0
asttokens==3.0.0
async-lru==2.0.4
attrs==24.3.0
audioread==3.0.1
babel==2.16.0
beautifulsoup4==4.12.3
bleach==6.2.0
blinker==1.9.0
celluloid==0.2.0
certifi==2024.12.14
cffi==1.17.1
charset-normalizer==3.4.1
click==8.1.8
comm==0.2.2
contourpy==1.3.1
cycler==0.12.1
debugpy==1.8.12
decorator==5.1.1
deepspeed==0.12.4
defusedxml==0.7.1
executing==2.2.0
fastapi==0.115.6
fastjsonschema==2.21.1
ffmpy==0.5.0
filelock==3.16.1
Flask==3.1.0
fonttools==4.55.3
fqdn==1.5.1
fsspec==2024.12.0
gradio==4.8.0
gradio_client==0.7.1
h11==0.14.0
hjson==3.1.0
httpcore==1.0.7
httpx==0.28.1
huggingface-hub==0.27.1
idna==3.10
importlib_resources==6.5.2
ipykernel==6.29.5
ipython==8.31.0
isoduration==20.11.0
itsdangerous==2.2.0
jedi==0.19.2
Jinja2==3.1.5
joblib==1.4.2
json5==0.10.0
jsonpointer==3.0.0
jsonschema==4.23.0
jsonschema-specifications==2024.10.1
jupyter-events==0.11.0
jupyter-lsp==2.2.5
jupyter_client==8.6.3
jupyter_core==5.7.2
jupyter_server==2.15.0
jupyter_server_terminals==0.5.3
jupyterlab==4.3.4
jupyterlab_pygments==0.3.0
jupyterlab_server==2.27.3
kiwisolver==1.4.8
lazy_loader==0.4
librosa==0.10.1
llvmlite==0.43.0
markdown-it-py==3.0.0
MarkupSafe==2.1.5
matplotlib==3.8.1
matplotlib-inline==0.1.7
mdurl==0.1.2
mistune==3.1.0
mpmath==1.3.0
msgpack==1.1.0
narwhals==1.22.0
nbclient==0.10.2
nbconvert==7.16.5
nbformat==5.10.4
nest-asyncio==1.6.0
networkx==3.4.2
ninja==1.11.1.3
notebook==7.3.2
notebook_shim==0.2.4
numba==0.60.0
numpy==1.26.2
nvidia-cublas-cu12==12.1.3.1
nvidia-cuda-cupti-cu12==12.1.105
nvidia-cuda-nvrtc-cu12==12.1.105
nvidia-cuda-runtime-cu12==12.1.105
nvidia-cudnn-cu12==8.9.2.26
nvidia-cufft-cu12==11.0.2.54
nvidia-curand-cu12==10.3.2.106
nvidia-cusolver-cu12==11.4.5.107
nvidia-cusparse-cu12==12.1.0.106
nvidia-ml-py==12.560.30
nvidia-nccl-cu12==2.18.1
nvidia-nvjitlink-cu12==12.6.85
nvidia-nvtx-cu12==12.1.105
omegaconf==2.3.0
orjson==3.10.14
overrides==7.7.0
packaging==24.2
pandas==2.1.3
pandocfilters==1.5.1
parso==0.8.4
pexpect==4.9.0
pillow==10.4.0
platformdirs==4.3.6
pooch==1.8.2
prometheus_client==0.21.1
prompt_toolkit==3.0.50
psutil==6.1.1
ptflops==0.7.1.2
ptyprocess==0.7.0
pure_eval==0.2.3
py-cpuinfo==9.0.0
pycparser==2.22
pydantic==2.10.5
pydantic_core==2.27.2
pydub==0.25.1
Pygments==2.19.1
pynvml==12.0.0
pyparsing==3.2.1
python-dateutil==2.9.0.post0
python-json-logger==3.2.1
python-multipart==0.0.20
pytz==2024.2
PyYAML==6.0.2
pyzmq==26.2.0
referencing==0.36.1
requests==2.32.3
resampy==0.4.2
resemble-enhance==0.0.1
rfc3339-validator==0.1.4
rfc3986-validator==0.1.1
rich==13.7.0
rpds-py==0.22.3
scikit-learn==1.6.1
scipy==1.11.4
semantic-version==2.10.0
Send2Trash==1.8.3
shellingham==1.5.4
six==1.17.0
sniffio==1.3.1
soundfile==0.12.1
soupsieve==2.6
soxr==0.5.0.post1
stack-data==0.6.3
starlette==0.41.3
sympy==1.13.3
tabulate==0.8.10
terminado==0.18.1
threadpoolctl==3.5.0
tinycss2==1.4.0
tomlkit==0.12.0
torch==2.1.1
torchaudio==2.1.1
torchvision==0.16.1
tornado==6.4.2
tqdm==4.66.1
traitlets==5.14.3
triton==2.1.0
typer==0.15.1
types-python-dateutil==2.9.0.20241206
typing_extensions==4.12.2
tzdata==2024.2
uri-template==1.3.0
urllib3==2.3.0
uvicorn==0.34.0
wcwidth==0.2.13
webcolors==24.11.1
webencodings==0.5.1
websocket-client==1.8.0
websockets==11.0.3
Werkzeug==3.1.3
"""

ORIGINAL_DEPENDENCIES = '''
resemble-enhance
'''


from typing import Iterable, Union, Type
from ...common import IDecider, Loc, Locator
from ..controller import IController, ControllerOverDecider
from ...deployment import Command, LocalExecutor
import json
from pathlib import Path
import importlib
import inspect
from dataclasses import dataclass


@dataclass
class InstallationStatus:
    installed: bool
    dockerless_controller: bool
    size: str|None = None


class ControllerRegistry:
    def __init__(self,
                 controllers: Iterable[Union[IController, IDecider, Type[IController], Type[IDecider]]],
                 locator: Locator = Loc
                 ):
        updated_controllers = []
        for i, c in enumerate(controllers):
            updated_controllers.append(ControllerRegistry.to_controller(c))
        self._controllers = tuple(updated_controllers)
        self._name_to_controller = {c.get_name(): c for c in self._controllers}
        self.locator = locator

    @staticmethod
    def to_controller_name(obj: str|IDecider|IController|type[IDecider]|type[IController]):
        if isinstance(obj, IDecider):
            return type(obj).__name__
        if isinstance(obj, IController):
            return obj.get_name()
        if isinstance(obj, str):
            return obj
        if isinstance(obj, type):
            if issubclass(obj, IDecider):
                return obj.__name__
            if issubclass(obj, IController):
                return obj.__name__.replace("Controller","")
        raise ValueError(f"Expected str or instance/type of IDecider, IController, but was {obj}")

    @staticmethod
    def to_controller(obj: IDecider|IController|type[IDecider]|type[IController]) -> IController:
        if isinstance(obj, IController):
            return obj
        if isinstance(obj, IDecider):
            tp = type(obj)
            if hasattr(tp, 'Controller'):
                return tp.Controller()
            return ControllerOverDecider(obj)
        if isinstance(obj, type):
            if issubclass(obj, IController):
                return obj()
            if issubclass(obj, IDecider):
                if hasattr(obj, 'Controller'):
                    return obj.Controller()
                return ControllerOverDecider(obj())
        raise ValueError(f"Expected instance or type of IDecider or IController, but was {obj}")

    def get_controller(self, name: str) -> IController:
        if name not in self._name_to_controller:
            raise ValueError(f"The requested controller {name} is not in the list of controllers")
        controller = self._name_to_controller[name]
        controller.context._resource_folder_root = self.locator.resources_folder
        return controller

    def get_deciders_names(self) -> list[str]:
        return list(self._name_to_controller)


    def get_installation_statuses(self) -> dict[str, InstallationStatus]:
        executor = LocalExecutor()
        command_output = executor.execute([
            'docker',
            'images',
            '--format',
            """{"repository": "{{.Repository}}", "size": "{{.Size}}"}"""
        ], Command.Options(return_output=True))

        images_data = []
        for s in command_output.split('\n'):
            if s.strip() == '':
                continue
            images_data.append(json.loads(s))

        images = {image['repository']: image for image in images_data}
        images_names = tuple(images)
        result = {}

        for name in self.get_deciders_names():
            container = self.get_controller(name)
            found_image_name = container.find_self_in_list(images_names)
            if found_image_name is None:
                result[name] = InstallationStatus(False, False)
                continue
            if isinstance(found_image_name, IController.DockerlessInstallation):
                result[name ] = InstallationStatus(True, True)
                continue
            if found_image_name not in images:
                raise ValueError(f"Controller breached contract on `find_self_in_list` method, returned {found_image_name}")
            result[name] = InstallationStatus(True, False, images[found_image_name]['size'])

        return result


    @staticmethod
    def discover() -> 'ControllerRegistry':
        module = importlib.import_module('brainbox.deciders')

        classes = [obj for _, obj in inspect.getmembers(module, inspect.isclass)]

        controllers = []
        for c in classes:
            if not isinstance(c, type):
                continue
            if not issubclass(c, IController) and not issubclass(c, IDecider):
                continue
            controllers.append(ControllerRegistry.to_controller(c))

        return ControllerRegistry(controllers)

    @staticmethod
    def discover_or_create(services: Iterable[IDecider|IController]|None):
        if services is None:
            return ControllerRegistry.discover()
        return ControllerRegistry(services)